##############################################################################
# Copyright (c) 2014 Hajime Nakagami<nakagami@gmail.com>
# All rights reserved.
# Licensed under the New BSD License
# (http://www.freebsd.org/copyright/freebsd-license.html)
#
# Python DB-API 2.0 module for Firebird. 
##############################################################################
# This SRP implementation is in reference to
'''
Following document was copied from <http://srp.stanford.edu/design.html>.
-----
SRP Protocol Design

SRP is the newest addition to a new class of strong authentication protocols that resist all the well-known passive and active attacks over the network. SRP borrows some elements from other key-exchange and identification protcols and adds some subtlee
 modifications and refinements. The result is a protocol that preserves the strength and efficiency of the EKE family protocols while fixing some of their shortcomings. 

The following is a description of SRP-6 and 6a, the latest versions of SRP: 

  N    A large safe prime (N = 2q+1, where q is prime)
       All arithmetic is done modulo N.
  g    A generator modulo N
  k    Multiplier parameter (k = H(N, g) in SRP-6a, k = 3 for legacy SRP-6)
  s    User's salt
  I    Username
  p    Cleartext Password
  H()  One-way hash function
  ^    (Modular) Exponentiation
  u    Random scrambling parameter
  a,b  Secret ephemeral values
  A,B  Public ephemeral values
  x    Private key (derived from p and s)
  v    Password verifier

The host stores passwords using the following formula: 

  x = H(s, p)               (s is chosen randomly)
  v = g^x                   (computes password verifier)

The host then keeps {I, s, v} in its password database. The authentication protocol itself goes as follows: 

User -> Host:  I, A = g^a                  (identifies self, a = random number)
Host -> User:  s, B = kv + g^b             (sends salt, b = random number)

        Both:  u = H(A, B)

        User:  x = H(s, p)                 (user enters password)
        User:  S = (B - kg^x) ^ (a + ux)   (computes session key)
        User:  K = H(S)

        Host:  S = (Av^u) ^ b              (computes session key)
        Host:  K = H(S)

Now the two parties have a shared, strong session key K. To complete authentication, they need to prove to each other that their keys match. One possible way: 

User -> Host:  M = H(H(N) xor H(g), H(I), s, A, B, K)
Host -> User:  H(A, M, K)

The two parties also employ the following safeguards: 

  1. The user will abort if he receives B == 0 (mod N) or u == 0. 
  2. The host will abort if it detects that A == 0 (mod N). 
  3. The user must show his proof of K first. If the server detects that the user's proof is incorrect, it must abort without showing its own proof of K. 

See http://srp.stanford.edu/ for more information.
'''
from __future__ import print_function
import sys
import hashlib
import hmac
import random
import binascii

DEBUG=False
DEBUG_PRINT=False
if DEBUG:
    DEBUG_PRIVATE_KEY = 0x60975527035CF2AD1989806F0407210BC81EDC04E2762A56AFD529DDDA2D4393

PYTHON_MAJOR_VER = sys.version_info[0]

if PYTHON_MAJOR_VER == 3:
    def ord(c):
        return c

SRP_KEY_SIZE = 128
SRP_SALT_SIZE = 32

def get_prime():
    N = 0xE67D2E994B2F900C3F41F08F5BB2627ED0D49EE1FE767A52EFCD565CD6E768812C3E1E9CE8F0A8BEA6CB13CD29DDEBF7A96D4A93B55D488DF099A15C89DCB0640738EB2CBDD9A8F7BAB561AB1B0DC1C6CDABF303264A08D1BCA932D1F1EE428B619D970F342ABA9A65793B8B2F041AE5364350C16F735F56ECBCA87BD57B29E7
    g = 2

    #n = N
    #scale = 0
    #while n > 0:
    #  scale += 1
    #  n >>= 8
    scale = 128

    #k = bytes2long(sha1(pad(N, scale), pad(g, scale)))
    k = 1277432915985975349439481660349303019122249719989

    return N, g, scale, k


def bytes2long(s):
    n = 0
    for c in s:
        n <<= 8
        n += ord(c)
    return n

def long2bytes(n):
    s = []
    while n > 0:
      s.insert(0, n & 255)
      n >>= 8
    if PYTHON_MAJOR_VER == 3:
        return bytes(s)
    else:
        return b''.join([chr(c) for c in s])

def sha1(*args):
    sha1 = hashlib.sha1()
    for v in args:
        if not isinstance(v, bytes):
            v = long2bytes(v)
        sha1.update(v)
    return sha1.digest()

def pad(n, scale):
    s = []
    for x in range(scale):
        s.insert(0, n & 255)
        n >>= 8
    if PYTHON_MAJOR_VER == 3:
        return bytes(s)
    else:
        return b''.join([chr(c) for c in s])

def get_scramble(x, y, scale):
    return bytes2long(sha1(pad(x, scale), pad(y, scale)))

def getUserHash(salt, user, password):
    assert isinstance(user, bytes)
    assert isinstance(password, bytes)
    hash1 = sha1(user, b':', password)
    hash2 = sha1(salt, hash1)
    rc =  bytes2long(hash2)

    return rc

def client_seed(user, password):
    """
        A: Client public key
        a: Client private key
    """
    N, g, scale, k  = get_prime()
    a = random.randrange(0, 1 << SRP_KEY_SIZE)
    A = pow(g, a, N)
    if DEBUG:
        a = DEBUG_PRIVATE_KEY
        A = pow(g, a, N)
    if DEBUG_PRINT:
        print('A=', binascii.b2a_hex(long2bytes(A)), end='\n')
        print('a=', binascii.b2a_hex(long2bytes(a)), end='\n')
    return A, a

def server_seed(v):
    """
        B: Server public key
        b: Server private key
    """
    N, g, scale, k = get_prime()
    b = random.randrange(0, 1 << SRP_KEY_SIZE)
    gb = pow(g, b, N)
    kv = (k * v) % N
    B = (kv + gb) % N
    if DEBUG:
        b = DEBUG_PRIVATE_KEY
        gb = pow(g, b, N)
        kv = (k * v) % N
        B = (kv + gb) % N
    if DEBUG_PRINT:
        print('B=', binascii.b2a_hex(long2bytes(B)), end='\n')
        print('b=', binascii.b2a_hex(long2bytes(b)), end='\n')
    return B, b

def client_session(user, password, salt, A, B, a):
    """
    Client session secret
        Both:  u = H(A, B)

        User:  x = H(s, p)                 (user enters password)
        User:  S = (B - kg^x) ^ (a + ux)   (computes session key)
        User:  K = H(S)
    """
    N, g, scale, k = get_prime()
    u = get_scramble(A, B, scale)
    x = getUserHash(salt, user, password)   # x
    gx = pow(g, x, N)                       # g^x
    kgx = (k * gx) % N                      # kg^x
    diff = (B - kgx) % N                    # B - kg^x
    ux = (u * x) % N
    aux = (a + ux) % N
    session_secret = pow(diff, aux, N)      # (B - kg^x) ^ (a + ux)
    K = sha1(session_secret)

    return K

def server_session(user, password, salt, A, B, b):
    """
    Server session secret
        Both:  u = H(A, B)

        Host:  S = (Av^u) ^ b              (computes session key)
        Host:  K = H(S)
    """
    N, g, scale, k = get_prime()
    u = get_scramble(A, B, scale)
    v = get_verifier(user, password, salt)
    vu = pow(v, u, N)                       # v^u
    Avu = (A * vu) % N                      # Av^u
    session_secret = pow(Avu, b, N)         # (Av^u) ^ b
    K = sha1(session_secret)
    if DEBUG_PRINT:
        print('server session_secret=',
            binascii.b2a_hex(long2bytes(session_secret)), end='\n')
        print('server session hash K=', binascii.b2a_hex(K))

    return K

def client_proof(user, password, salt, A, B, a):
    """
    M = H(H(N) xor H(g), H(I), s, A, B, K)
    """
    N, g, scale, k = get_prime()
    K = client_session(user, password, salt, A, B, a)

    n1 = bytes2long(sha1(N))
    n2 = bytes2long(sha1(g))
    n1 = pow(n1, n2, N)
    n2 = bytes2long(sha1(user))
    M = sha1(n1, n2, salt, A, B, K)
    if DEBUG_PRINT:
        print('client_proof:M=', binascii.b2a_hex(M), end='\n')
        print('client_proof:K=', binascii.b2a_hex(K), end='\n')

    return M, K

def get_salt():
    if PYTHON_MAJOR_VER == 3:
        return bytes([random.randrange(0, 256) for x in range(SRP_SALT_SIZE)])
    else:
        return b''.join([chr(random.randrange(0, 256)) for x in range(SRP_SALT_SIZE)])

def get_verifier(user, password, salt):
    N, g, scale, k = get_prime()
    x = getUserHash(salt, user, password)
    return pow(g, x, N)

if __name__ == '__main__':
    """
    A, a, B, b are long.
    salt, M are bytes.
    """
    # Both
    user = b'SYSDBA'
    password = b'masterkey'

    # Client send A to Server
    A, a = client_seed(user, password)

    # Server send B, salt to Client
    salt = get_salt()
    v = get_verifier(user, password, salt)
    B, b = server_seed(v)

    serverKey = server_session(user, password, salt, A, B, b)

    # Client send M to Server
    M, clientKey = client_proof(user, password, salt, A, B, a)

    # Client and Server has same key
    assert clientKey == serverKey




#############################################################################
# The contents of this file are subject to the Interbase Public
# License Version 1.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy
# of the License at http://www.Inprise.com/IPL.html
# 
# Software distributed under the License is distributed on an
# "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
# or implied. See the License for the specific language governing
# rights and limitations under the License.

messages = {
    335544321 : '''arithmetic exception, numeric overflow, or string truncation\n''', 
    335544322 : '''invalid database key\n''', 
    335544323 : '''file @1 is not a valid database\n''', 
    335544324 : '''invalid database handle (no active connection)\n''', 
    335544325 : '''bad parameters on attach or create database\n''', 
    335544326 : '''unrecognized database parameter block\n''', 
    335544327 : '''invalid request handle\n''', 
    335544328 : '''invalid BLOB handle\n''', 
    335544329 : '''invalid BLOB ID\n''', 
    335544330 : '''invalid parameter in transaction parameter block\n''', 
    335544331 : '''invalid format for transaction parameter block\n''', 
    335544332 : '''invalid transaction handle (expecting explicit transaction start)\n''', 
    335544333 : '''internal Firebird consistency check (@1)\n''', 
    335544334 : '''conversion error from string "@1"\n''', 
    335544335 : '''database file appears corrupt (@1)\n''', 
    335544336 : '''deadlock\n''', 
    335544337 : '''attempt to start more than @1 transactions\n''', 
    335544338 : '''no match for first value expression\n''', 
    335544339 : '''information type inappropriate for object specified\n''', 
    335544340 : '''no information of this type available for object specified\n''', 
    335544341 : '''unknown information item\n''', 
    335544342 : '''action cancelled by trigger (@1) to preserve data integrity\n''', 
    335544343 : '''invalid request BLR at offset @1\n''', 
    335544344 : '''I/O error during "@1" operation for file "@2"\n''', 
    335544345 : '''lock conflict on no wait transaction\n''', 
    335544346 : '''corrupt system table\n''', 
    335544347 : '''validation error for column @1, value "@2"\n''', 
    335544348 : '''no current record for fetch operation\n''', 
    335544349 : '''attempt to store duplicate value (visible to active transactions) in unique index "@1"\n''', 
    335544350 : '''program attempted to exit without finishing database\n''', 
    335544351 : '''unsuccessful metadata update\n''', 
    335544352 : '''no permission for @1 access to @2 @3\n''', 
    335544353 : '''transaction is not in limbo\n''', 
    335544354 : '''invalid database key\n''', 
    335544355 : '''BLOB was not closed\n''', 
    335544356 : '''metadata is obsolete\n''', 
    335544357 : '''cannot disconnect database with open transactions (@1 active)\n''', 
    335544358 : '''message length error (encountered @1, expected @2)\n''', 
    335544359 : '''attempted update of read-only column\n''', 
    335544360 : '''attempted update of read-only table\n''', 
    335544361 : '''attempted update during read-only transaction\n''', 
    335544362 : '''cannot update read-only view @1\n''', 
    335544363 : '''no transaction for request\n''', 
    335544364 : '''request synchronization error\n''', 
    335544365 : '''request referenced an unavailable database\n''', 
    335544366 : '''segment buffer length shorter than expected\n''', 
    335544367 : '''attempted retrieval of more segments than exist\n''', 
    335544368 : '''attempted invalid operation on a BLOB\n''', 
    335544369 : '''attempted read of a new, open BLOB\n''', 
    335544370 : '''attempted action on BLOB outside transaction\n''', 
    335544371 : '''attempted write to read-only BLOB\n''', 
    335544372 : '''attempted reference to BLOB in unavailable database\n''', 
    335544373 : '''operating system directive @1 failed\n''', 
    335544374 : '''attempt to fetch past the last record in a record stream\n''', 
    335544375 : '''unavailable database\n''', 
    335544376 : '''table @1 was omitted from the transaction reserving list\n''', 
    335544377 : '''request includes a DSRI extension not supported in this implementation\n''', 
    335544378 : '''feature is not supported\n''', 
    335544379 : '''unsupported on-disk structure for file @1; found @2.@3, support @4.@5\n''', 
    335544380 : '''wrong number of arguments on call\n''', 
    335544381 : '''Implementation limit exceeded\n''', 
    335544382 : '''@1\n''', 
    335544383 : '''unrecoverable conflict with limbo transaction @1\n''', 
    335544384 : '''internal error\n''', 
    335544385 : '''internal error\n''', 
    335544386 : '''too many requests\n''', 
    335544387 : '''internal error\n''', 
    335544388 : '''block size exceeds implementation restriction\n''', 
    335544389 : '''buffer exhausted\n''', 
    335544390 : '''BLR syntax error: expected @1 at offset @2, encountered @3\n''', 
    335544391 : '''buffer in use\n''', 
    335544392 : '''internal error\n''', 
    335544393 : '''request in use\n''', 
    335544394 : '''incompatible version of on-disk structure\n''', 
    335544395 : '''table @1 is not defined\n''', 
    335544396 : '''column @1 is not defined in table @2\n''', 
    335544397 : '''internal error\n''', 
    335544398 : '''internal error\n''', 
    335544399 : '''internal error\n''', 
    335544400 : '''internal error\n''', 
    335544401 : '''internal error\n''', 
    335544402 : '''internal error\n''', 
    335544403 : '''page @1 is of wrong type (expected @2, found @3)\n''', 
    335544404 : '''database corrupted\n''', 
    335544405 : '''checksum error on database page @1\n''', 
    335544406 : '''index is broken\n''', 
    335544407 : '''database handle not zero\n''', 
    335544408 : '''transaction handle not zero\n''', 
    335544409 : '''transaction--request mismatch (synchronization error)\n''', 
    335544410 : '''bad handle count\n''', 
    335544411 : '''wrong version of transaction parameter block\n''', 
    335544412 : '''unsupported BLR version (expected @1, encountered @2)\n''', 
    335544413 : '''wrong version of database parameter block\n''', 
    335544414 : '''BLOB and array data types are not supported for @1 operation\n''', 
    335544415 : '''database corrupted\n''', 
    335544416 : '''internal error\n''', 
    335544417 : '''internal error\n''', 
    335544418 : '''transaction in limbo\n''', 
    335544419 : '''transaction not in limbo\n''', 
    335544420 : '''transaction outstanding\n''', 
    335544421 : '''connection rejected by remote interface\n''', 
    335544422 : '''internal error\n''', 
    335544423 : '''internal error\n''', 
    335544424 : '''no lock manager available\n''', 
    335544425 : '''context already in use (BLR error)\n''', 
    335544426 : '''context not defined (BLR error)\n''', 
    335544427 : '''data operation not supported\n''', 
    335544428 : '''undefined message number\n''', 
    335544429 : '''undefined parameter number\n''', 
    335544430 : '''unable to allocate memory from operating system\n''', 
    335544431 : '''blocking signal has been received\n''', 
    335544432 : '''lock manager error\n''', 
    335544433 : '''communication error with journal "@1"\n''', 
    335544434 : '''key size exceeds implementation restriction for index "@1"\n''', 
    335544435 : '''null segment of UNIQUE KEY\n''', 
    335544436 : '''SQL error code = @1\n''', 
    335544437 : '''wrong DYN version\n''', 
    335544438 : '''function @1 is not defined\n''', 
    335544439 : '''function @1 could not be matched\n''', 
    335544440 : '''\n''', 
    335544441 : '''database detach completed with errors\n''', 
    335544442 : '''database system cannot read argument @1\n''', 
    335544443 : '''database system cannot write argument @1\n''', 
    335544444 : '''operation not supported\n''', 
    335544445 : '''@1 extension error\n''', 
    335544446 : '''not updatable\n''', 
    335544447 : '''no rollback performed\n''', 
    335544448 : '''\n''', 
    335544449 : '''\n''', 
    335544450 : '''@1\n''', 
    335544451 : '''update conflicts with concurrent update\n''', 
    335544452 : '''product @1 is not licensed\n''', 
    335544453 : '''object @1 is in use\n''', 
    335544454 : '''filter not found to convert type @1 to type @2\n''', 
    335544455 : '''cannot attach active shadow file\n''', 
    335544456 : '''invalid slice description language at offset @1\n''', 
    335544457 : '''subscript out of bounds\n''', 
    335544458 : '''column not array or invalid dimensions (expected @1, encountered @2)\n''', 
    335544459 : '''record from transaction @1 is stuck in limbo\n''', 
    335544460 : '''a file in manual shadow @1 is unavailable\n''', 
    335544461 : '''secondary server attachments cannot validate databases\n''', 
    335544462 : '''secondary server attachments cannot start journaling\n''', 
    335544463 : '''generator @1 is not defined\n''', 
    335544464 : '''secondary server attachments cannot start logging\n''', 
    335544465 : '''invalid BLOB type for operation\n''', 
    335544466 : '''violation of FOREIGN KEY constraint "@1" on table "@2"\n''', 
    335544467 : '''minor version too high found @1 expected @2\n''', 
    335544468 : '''transaction @1 is @2\n''', 
    335544469 : '''transaction marked invalid by I/O error\n''', 
    335544470 : '''cache buffer for page @1 invalid\n''', 
    335544471 : '''there is no index in table @1 with id @2\n''', 
    335544472 : '''Your user name and password are not defined. Ask your database administrator to set up a Firebird login.\n''', 
    335544473 : '''invalid bookmark handle\n''', 
    335544474 : '''invalid lock level @1\n''', 
    335544475 : '''lock on table @1 conflicts with existing lock\n''', 
    335544476 : '''requested record lock conflicts with existing lock\n''', 
    335544477 : '''maximum indexes per table (@1) exceeded\n''', 
    335544478 : '''enable journal for database before starting online dump\n''', 
    335544479 : '''online dump failure. Retry dump\n''', 
    335544480 : '''an online dump is already in progress\n''', 
    335544481 : '''no more disk/tape space.  Cannot continue online dump\n''', 
    335544482 : '''journaling allowed only if database has Write-ahead Log\n''', 
    335544483 : '''maximum number of online dump files that can be specified is 16\n''', 
    335544484 : '''error in opening Write-ahead Log file during recovery\n''', 
    335544485 : '''invalid statement handle\n''', 
    335544486 : '''Write-ahead log subsystem failure\n''', 
    335544487 : '''WAL Writer error\n''', 
    335544488 : '''Log file header of @1 too small\n''', 
    335544489 : '''Invalid version of log file @1\n''', 
    335544490 : '''Log file @1 not latest in the chain but open flag still set\n''', 
    335544491 : '''Log file @1 not closed properly; database recovery may be required\n''', 
    335544492 : '''Database name in the log file @1 is different\n''', 
    335544493 : '''Unexpected end of log file @1 at offset @2\n''', 
    335544494 : '''Incomplete log record at offset @1 in log file @2\n''', 
    335544495 : '''Log record header too small at offset @1 in log file @2\n''', 
    335544496 : '''Log block too small at offset @1 in log file @2\n''', 
    335544497 : '''Illegal attempt to attach to an uninitialized WAL segment for @1\n''', 
    335544498 : '''Invalid WAL parameter block option @1\n''', 
    335544499 : '''Cannot roll over to the next log file @1\n''', 
    335544500 : '''database does not use Write-ahead Log\n''', 
    335544501 : '''cannot drop log file when journaling is enabled\n''', 
    335544502 : '''reference to invalid stream number\n''', 
    335544503 : '''WAL subsystem encountered error\n''', 
    335544504 : '''WAL subsystem corrupted\n''', 
    335544505 : '''must specify archive file when enabling long term journal for databases with round-robin log files\n''', 
    335544506 : '''database @1 shutdown in progress\n''', 
    335544507 : '''refresh range number @1 already in use\n''', 
    335544508 : '''refresh range number @1 not found\n''', 
    335544509 : '''CHARACTER SET @1 is not defined\n''', 
    335544510 : '''lock time-out on wait transaction\n''', 
    335544511 : '''procedure @1 is not defined\n''', 
    335544512 : '''Input parameter mismatch for procedure @1\n''', 
    335544513 : '''Database @1: WAL subsystem bug for pid @2@3\n''', 
    335544514 : '''Could not expand the WAL segment for database @1\n''', 
    335544515 : '''status code @1 unknown\n''', 
    335544516 : '''exception @1 not defined\n''', 
    335544517 : '''exception @1\n''', 
    335544518 : '''restart shared cache manager\n''', 
    335544519 : '''invalid lock handle\n''', 
    335544520 : '''long-term journaling already enabled\n''', 
    335544521 : '''Unable to roll over please see Firebird log.\n''', 
    335544522 : '''WAL I/O error.  Please see Firebird log.\n''', 
    335544523 : '''WAL writer - Journal server communication error.  Please see Firebird log.\n''', 
    335544524 : '''WAL buffers cannot be increased.  Please see Firebird log.\n''', 
    335544525 : '''WAL setup error.  Please see Firebird log.\n''', 
    335544526 : '''obsolete\n''', 
    335544527 : '''Cannot start WAL writer for the database @1\n''', 
    335544528 : '''database @1 shutdown\n''', 
    335544529 : '''cannot modify an existing user privilege\n''', 
    335544530 : '''Cannot delete PRIMARY KEY being used in FOREIGN KEY definition.\n''', 
    335544531 : '''Column used in a PRIMARY constraint must be NOT NULL.\n''', 
    335544532 : '''Name of Referential Constraint not defined in constraints table.\n''', 
    335544533 : '''Non-existent PRIMARY or UNIQUE KEY specified for FOREIGN KEY.\n''', 
    335544534 : '''Cannot update constraints (RDB$REF_CONSTRAINTS).\n''', 
    335544535 : '''Cannot update constraints (RDB$CHECK_CONSTRAINTS).\n''', 
    335544536 : '''Cannot delete CHECK constraint entry (RDB$CHECK_CONSTRAINTS)\n''', 
    335544537 : '''Cannot delete index segment used by an Integrity Constraint\n''', 
    335544538 : '''Cannot update index segment used by an Integrity Constraint\n''', 
    335544539 : '''Cannot delete index used by an Integrity Constraint\n''', 
    335544540 : '''Cannot modify index used by an Integrity Constraint\n''', 
    335544541 : '''Cannot delete trigger used by a CHECK Constraint\n''', 
    335544542 : '''Cannot update trigger used by a CHECK Constraint\n''', 
    335544543 : '''Cannot delete column being used in an Integrity Constraint.\n''', 
    335544544 : '''Cannot rename column being used in an Integrity Constraint.\n''', 
    335544545 : '''Cannot update constraints (RDB$RELATION_CONSTRAINTS).\n''', 
    335544546 : '''Cannot define constraints on views\n''', 
    335544547 : '''internal Firebird consistency check (invalid RDB$CONSTRAINT_TYPE)\n''', 
    335544548 : '''Attempt to define a second PRIMARY KEY for the same table\n''', 
    335544549 : '''cannot modify or erase a system trigger\n''', 
    335544550 : '''only the owner of a table may reassign ownership\n''', 
    335544551 : '''could not find object for GRANT\n''', 
    335544552 : '''could not find column for GRANT\n''', 
    335544553 : '''user does not have GRANT privileges for operation\n''', 
    335544554 : '''object has non-SQL security class defined\n''', 
    335544555 : '''column has non-SQL security class defined\n''', 
    335544556 : '''Write-ahead Log without shared cache configuration not allowed\n''', 
    335544557 : '''database shutdown unsuccessful\n''', 
    335544558 : '''Operation violates CHECK constraint @1 on view or table @2\n''', 
    335544559 : '''invalid service handle\n''', 
    335544560 : '''database @1 shutdown in @2 seconds\n''', 
    335544561 : '''wrong version of service parameter block\n''', 
    335544562 : '''unrecognized service parameter block\n''', 
    335544563 : '''service @1 is not defined\n''', 
    335544564 : '''long-term journaling not enabled\n''', 
    335544565 : '''Cannot transliterate character between character sets\n''', 
    335544566 : '''WAL defined; Cache Manager must be started first\n''', 
    335544567 : '''Overflow log specification required for round-robin log\n''', 
    335544568 : '''Implementation of text subtype @1 not located.\n''', 
    335544569 : '''Dynamic SQL Error\n''', 
    335544570 : '''Invalid command\n''', 
    335544571 : '''Data type for constant unknown\n''', 
    335544572 : '''Invalid cursor reference\n''', 
    335544573 : '''Data type unknown\n''', 
    335544574 : '''Invalid cursor declaration\n''', 
    335544575 : '''Cursor @1 is not updatable\n''', 
    335544576 : '''Attempt to reopen an open cursor\n''', 
    335544577 : '''Attempt to reclose a closed cursor\n''', 
    335544578 : '''Column unknown\n''', 
    335544579 : '''Internal error\n''', 
    335544580 : '''Table unknown\n''', 
    335544581 : '''Procedure unknown\n''', 
    335544582 : '''Request unknown\n''', 
    335544583 : '''SQLDA missing or incorrect version, or incorrect number/type of variables\n''', 
    335544584 : '''Count of read-write columns does not equal count of values\n''', 
    335544585 : '''Invalid statement handle\n''', 
    335544586 : '''Function unknown\n''', 
    335544587 : '''Column is not a BLOB\n''', 
    335544588 : '''COLLATION @1 for CHARACTER SET @2 is not defined\n''', 
    335544589 : '''COLLATION @1 is not valid for specified CHARACTER SET\n''', 
    335544590 : '''Option specified more than once\n''', 
    335544591 : '''Unknown transaction option\n''', 
    335544592 : '''Invalid array reference\n''', 
    335544593 : '''Array declared with too many dimensions\n''', 
    335544594 : '''Illegal array dimension range\n''', 
    335544595 : '''Trigger unknown\n''', 
    335544596 : '''Subselect illegal in this context\n''', 
    335544597 : '''Cannot prepare a CREATE DATABASE/SCHEMA statement\n''', 
    335544598 : '''must specify column name for view select expression\n''', 
    335544599 : '''number of columns does not match select list\n''', 
    335544600 : '''Only simple column names permitted for VIEW WITH CHECK OPTION\n''', 
    335544601 : '''No WHERE clause for VIEW WITH CHECK OPTION\n''', 
    335544602 : '''Only one table allowed for VIEW WITH CHECK OPTION\n''', 
    335544603 : '''DISTINCT, GROUP or HAVING not permitted for VIEW WITH CHECK OPTION\n''', 
    335544604 : '''FOREIGN KEY column count does not match PRIMARY KEY\n''', 
    335544605 : '''No subqueries permitted for VIEW WITH CHECK OPTION\n''', 
    335544606 : '''expression evaluation not supported\n''', 
    335544607 : '''gen.c: node not supported\n''', 
    335544608 : '''Unexpected end of command\n''', 
    335544609 : '''INDEX @1\n''', 
    335544610 : '''EXCEPTION @1\n''', 
    335544611 : '''COLUMN @1\n''', 
    335544612 : '''Token unknown\n''', 
    335544613 : '''union not supported\n''', 
    335544614 : '''Unsupported DSQL construct\n''', 
    335544615 : '''column used with aggregate\n''', 
    335544616 : '''invalid column reference\n''', 
    335544617 : '''invalid ORDER BY clause\n''', 
    335544618 : '''Return mode by value not allowed for this data type\n''', 
    335544619 : '''External functions cannot have more than 10 parameters\n''', 
    335544620 : '''alias @1 conflicts with an alias in the same statement\n''', 
    335544621 : '''alias @1 conflicts with a procedure in the same statement\n''', 
    335544622 : '''alias @1 conflicts with a table in the same statement\n''', 
    335544623 : '''Illegal use of keyword VALUE\n''', 
    335544624 : '''segment count of 0 defined for index @1\n''', 
    335544625 : '''A node name is not permitted in a secondary, shadow, cache or log file name\n''', 
    335544626 : '''TABLE @1\n''', 
    335544627 : '''PROCEDURE @1\n''', 
    335544628 : '''cannot create index @1\n''', 
    335544629 : '''Write-ahead Log with shadowing configuration not allowed\n''', 
    335544630 : '''there are @1 dependencies\n''', 
    335544631 : '''too many keys defined for index @1\n''', 
    335544632 : '''Preceding file did not specify length, so @1 must include starting page number\n''', 
    335544633 : '''Shadow number must be a positive integer\n''', 
    335544634 : '''Token unknown - line @1, column @2\n''', 
    335544635 : '''there is no alias or table named @1 at this scope level\n''', 
    335544636 : '''there is no index @1 for table @2\n''', 
    335544637 : '''table @1 is not referenced in plan\n''', 
    335544638 : '''table @1 is referenced more than once in plan; use aliases to distinguish\n''', 
    335544639 : '''table @1 is referenced in the plan but not the from list\n''', 
    335544640 : '''Invalid use of CHARACTER SET or COLLATE\n''', 
    335544641 : '''Specified domain or source column @1 does not exist\n''', 
    335544642 : '''index @1 cannot be used in the specified plan\n''', 
    335544643 : '''the table @1 is referenced twice; use aliases to differentiate\n''', 
    335544644 : '''attempt to fetch before the first record in a record stream\n''', 
    335544645 : '''the current position is on a crack\n''', 
    335544646 : '''database or file exists\n''', 
    335544647 : '''invalid comparison operator for find operation\n''', 
    335544648 : '''Connection lost to pipe server\n''', 
    335544649 : '''bad checksum\n''', 
    335544650 : '''wrong page type\n''', 
    335544651 : '''Cannot insert because the file is readonly or is on a read only medium.\n''', 
    335544652 : '''multiple rows in singleton select\n''', 
    335544653 : '''cannot attach to password database\n''', 
    335544654 : '''cannot start transaction for password database\n''', 
    335544655 : '''invalid direction for find operation\n''', 
    335544656 : '''variable @1 conflicts with parameter in same procedure\n''', 
    335544657 : '''Array/BLOB/DATE data types not allowed in arithmetic\n''', 
    335544658 : '''@1 is not a valid base table of the specified view\n''', 
    335544659 : '''table @1 is referenced twice in view; use an alias to distinguish\n''', 
    335544660 : '''view @1 has more than one base table; use aliases to distinguish\n''', 
    335544661 : '''cannot add index, index root page is full.\n''', 
    335544662 : '''BLOB SUB_TYPE @1 is not defined\n''', 
    335544663 : '''Too many concurrent executions of the same request\n''', 
    335544664 : '''duplicate specification of @1 - not supported\n''', 
    335544665 : '''violation of PRIMARY or UNIQUE KEY constraint "@1" on table "@2"\n''', 
    335544666 : '''server version too old to support all CREATE DATABASE options\n''', 
    335544667 : '''drop database completed with errors\n''', 
    335544668 : '''procedure @1 does not return any values\n''', 
    335544669 : '''count of column list and variable list do not match\n''', 
    335544670 : '''attempt to index BLOB column in index @1\n''', 
    335544671 : '''attempt to index array column in index @1\n''', 
    335544672 : '''too few key columns found for index @1 (incorrect column name?)\n''', 
    335544673 : '''cannot delete\n''', 
    335544674 : '''last column in a table cannot be deleted\n''', 
    335544675 : '''sort error\n''', 
    335544676 : '''sort error: not enough memory\n''', 
    335544677 : '''too many versions\n''', 
    335544678 : '''invalid key position\n''', 
    335544679 : '''segments not allowed in expression index @1\n''', 
    335544680 : '''sort error: corruption in data structure\n''', 
    335544681 : '''new record size of @1 bytes is too big\n''', 
    335544682 : '''Inappropriate self-reference of column\n''', 
    335544683 : '''request depth exceeded. (Recursive definition?)\n''', 
    335544684 : '''cannot access column @1 in view @2\n''', 
    335544685 : '''dbkey not available for multi-table views\n''', 
    335544686 : '''journal file wrong format\n''', 
    335544687 : '''intermediate journal file full\n''', 
    335544688 : '''The prepare statement identifies a prepare statement with an open cursor\n''', 
    335544689 : '''Firebird error\n''', 
    335544690 : '''Cache redefined\n''', 
    335544691 : '''Insufficient memory to allocate page buffer cache\n''', 
    335544692 : '''Log redefined\n''', 
    335544693 : '''Log size too small\n''', 
    335544694 : '''Log partition size too small\n''', 
    335544695 : '''Partitions not supported in series of log file specification\n''', 
    335544696 : '''Total length of a partitioned log must be specified\n''', 
    335544697 : '''Precision must be from 1 to 18\n''', 
    335544698 : '''Scale must be between zero and precision\n''', 
    335544699 : '''Short integer expected\n''', 
    335544700 : '''Long integer expected\n''', 
    335544701 : '''Unsigned short integer expected\n''', 
    335544702 : '''Invalid ESCAPE sequence\n''', 
    335544703 : '''service @1 does not have an associated executable\n''', 
    335544704 : '''Failed to locate host machine.\n''', 
    335544705 : '''Undefined service @1/@2.\n''', 
    335544706 : '''The specified name was not found in the hosts file or Domain Name Services.\n''', 
    335544707 : '''user does not have GRANT privileges on base table/view for operation\n''', 
    335544708 : '''Ambiguous column reference.\n''', 
    335544709 : '''Invalid aggregate reference\n''', 
    335544710 : '''navigational stream @1 references a view with more than one base table\n''', 
    335544711 : '''Attempt to execute an unprepared dynamic SQL statement.\n''', 
    335544712 : '''Positive value expected\n''', 
    335544713 : '''Incorrect values within SQLDA structure\n''', 
    335544714 : '''invalid blob id\n''', 
    335544715 : '''Operation not supported for EXTERNAL FILE table @1\n''', 
    335544716 : '''Service is currently busy: @1\n''', 
    335544717 : '''stack size insufficent to execute current request\n''', 
    335544718 : '''Invalid key for find operation\n''', 
    335544719 : '''Error initializing the network software.\n''', 
    335544720 : '''Unable to load required library @1.\n''', 
    335544721 : '''Unable to complete network request to host "@1".\n''', 
    335544722 : '''Failed to establish a connection.\n''', 
    335544723 : '''Error while listening for an incoming connection.\n''', 
    335544724 : '''Failed to establish a secondary connection for event processing.\n''', 
    335544725 : '''Error while listening for an incoming event connection request.\n''', 
    335544726 : '''Error reading data from the connection.\n''', 
    335544727 : '''Error writing data to the connection.\n''', 
    335544728 : '''Cannot deactivate index used by an integrity constraint\n''', 
    335544729 : '''Cannot deactivate index used by a PRIMARY/UNIQUE constraint\n''', 
    335544730 : '''Client/Server Express not supported in this release\n''', 
    335544731 : '''\n''', 
    335544732 : '''Access to databases on file servers is not supported.\n''', 
    335544733 : '''Error while trying to create file\n''', 
    335544734 : '''Error while trying to open file\n''', 
    335544735 : '''Error while trying to close file\n''', 
    335544736 : '''Error while trying to read from file\n''', 
    335544737 : '''Error while trying to write to file\n''', 
    335544738 : '''Error while trying to delete file\n''', 
    335544739 : '''Error while trying to access file\n''', 
    335544740 : '''A fatal exception occurred during the execution of a user defined function.\n''', 
    335544741 : '''connection lost to database\n''', 
    335544742 : '''User cannot write to RDB$USER_PRIVILEGES\n''', 
    335544743 : '''token size exceeds limit\n''', 
    335544744 : '''Maximum user count exceeded.  Contact your database administrator.\n''', 
    335544745 : '''Your login @1 is same as one of the SQL role name. Ask your database administrator to set up a valid Firebird login.\n''', 
    335544746 : '''"REFERENCES table" without "(column)" requires PRIMARY KEY on referenced table\n''', 
    335544747 : '''The username entered is too long.  Maximum length is 31 bytes.\n''', 
    335544748 : '''The password specified is too long.  Maximum length is 8 bytes.\n''', 
    335544749 : '''A username is required for this operation.\n''', 
    335544750 : '''A password is required for this operation\n''', 
    335544751 : '''The network protocol specified is invalid\n''', 
    335544752 : '''A duplicate user name was found in the security database\n''', 
    335544753 : '''The user name specified was not found in the security database\n''', 
    335544754 : '''An error occurred while attempting to add the user.\n''', 
    335544755 : '''An error occurred while attempting to modify the user record.\n''', 
    335544756 : '''An error occurred while attempting to delete the user record.\n''', 
    335544757 : '''An error occurred while updating the security database.\n''', 
    335544758 : '''sort record size of @1 bytes is too big\n''', 
    335544759 : '''can not define a not null column with NULL as default value\n''', 
    335544760 : '''invalid clause --- '@1'\n''', 
    335544761 : '''too many open handles to database\n''', 
    335544762 : '''size of optimizer block exceeded\n''', 
    335544763 : '''a string constant is delimited by double quotes\n''', 
    335544764 : '''DATE must be changed to TIMESTAMP\n''', 
    335544765 : '''attempted update on read-only database\n''', 
    335544766 : '''SQL dialect @1 is not supported in this database\n''', 
    335544767 : '''A fatal exception occurred during the execution of a blob filter.\n''', 
    335544768 : '''Access violation.  The code attempted to access a virtual address without privilege to do so.\n''', 
    335544769 : '''Datatype misalignment.  The attempted to read or write a value that was not stored on a memory boundary.\n''', 
    335544770 : '''Array bounds exceeded.  The code attempted to access an array element that is out of bounds.\n''', 
    335544771 : '''Float denormal operand.  One of the floating-point operands is too small to represent a standard float value.\n''', 
    335544772 : '''Floating-point divide by zero.  The code attempted to divide a floating-point value by zero.\n''', 
    335544773 : '''Floating-point inexact result.  The result of a floating-point operation cannot be represented as a deciaml fraction.\n''', 
    335544774 : '''Floating-point invalid operand.  An indeterminant error occurred during a floating-point operation.\n''', 
    335544775 : '''Floating-point overflow.  The exponent of a floating-point operation is greater than the magnitude allowed.\n''', 
    335544776 : '''Floating-point stack check.  The stack overflowed or underflowed as the result of a floating-point operation.\n''', 
    335544777 : '''Floating-point underflow.  The exponent of a floating-point operation is less than the magnitude allowed.\n''', 
    335544778 : '''Integer divide by zero.  The code attempted to divide an integer value by an integer divisor of zero.\n''', 
    335544779 : '''Integer overflow.  The result of an integer operation caused the most significant bit of the result to carry.\n''', 
    335544780 : '''An exception occurred that does not have a description.  Exception number @1.\n''', 
    335544781 : '''Stack overflow.  The resource requirements of the runtime stack have exceeded the memory available to it.\n''', 
    335544782 : '''Segmentation Fault. The code attempted to access memory without priviledges.\n''', 
    335544783 : '''Illegal Instruction. The Code attempted to perfrom an illegal operation.\n''', 
    335544784 : '''Bus Error. The Code caused a system bus error.\n''', 
    335544785 : '''Floating Point Error. The Code caused an Arithmetic Exception or a floating point exception.\n''', 
    335544786 : '''Cannot delete rows from external files.\n''', 
    335544787 : '''Cannot update rows in external files.\n''', 
    335544788 : '''Unable to perform operation.  You must be either SYSDBA or owner of the database\n''', 
    335544789 : '''Specified EXTRACT part does not exist in input datatype\n''', 
    335544790 : '''Service @1 requires SYSDBA permissions.  Reattach to the Service Manager using the SYSDBA account.\n''', 
    335544791 : '''The file @1 is currently in use by another process.  Try again later.\n''', 
    335544792 : '''Cannot attach to services manager\n''', 
    335544793 : '''Metadata update statement is not allowed by the current database SQL dialect @1\n''', 
    335544794 : '''operation was cancelled\n''', 
    335544795 : '''unexpected item in service parameter block, expected @1\n''', 
    335544796 : '''Client SQL dialect @1 does not support reference to @2 datatype\n''', 
    335544797 : '''user name and password are required while attaching to the services manager\n''', 
    335544798 : '''You created an indirect dependency on uncommitted metadata. You must roll back the current transaction.\n''', 
    335544799 : '''The service name was not specified.\n''', 
    335544800 : '''Too many Contexts of Relation/Procedure/Views. Maximum allowed is 256\n''', 
    335544801 : '''data type not supported for arithmetic\n''', 
    335544802 : '''Database dialect being changed from 3 to 1\n''', 
    335544803 : '''Database dialect not changed.\n''', 
    335544804 : '''Unable to create database @1\n''', 
    335544805 : '''Database dialect @1 is not a valid dialect.\n''', 
    335544806 : '''Valid database dialects are @1.\n''', 
    335544807 : '''SQL warning code = @1\n''', 
    335544808 : '''DATE data type is now called TIMESTAMP\n''', 
    335544809 : '''Function @1 is in @2, which is not in a permitted directory for external functions.\n''', 
    335544810 : '''value exceeds the range for valid dates\n''', 
    335544811 : '''passed client dialect @1 is not a valid dialect.\n''', 
    335544812 : '''Valid client dialects are @1.\n''', 
    335544813 : '''Unsupported field type specified in BETWEEN predicate.\n''', 
    335544814 : '''Services functionality will be supported in a later version  of the product\n''', 
    335544815 : '''GENERATOR @1\n''', 
    335544816 : '''UDF @1\n''', 
    335544817 : '''Invalid parameter to FIRST.  Only integers >= 0 are allowed.\n''', 
    335544818 : '''Invalid parameter to SKIP.  Only integers >= 0 are allowed.\n''', 
    335544819 : '''File exceeded maximum size of 2GB.  Add another database file or use a 64 bit I/O version of Firebird.\n''', 
    335544820 : '''Unable to find savepoint with name @1 in transaction context\n''', 
    335544821 : '''Invalid column position used in the @1 clause\n''', 
    335544822 : '''Cannot use an aggregate or window function in a WHERE clause, use HAVING (for aggregate only) instead\n''', 
    335544823 : '''Cannot use an aggregate or window function in a GROUP BY clause\n''', 
    335544824 : '''Invalid expression in the @1 (not contained in either an aggregate function or the GROUP BY clause)\n''', 
    335544825 : '''Invalid expression in the @1 (neither an aggregate function nor a part of the GROUP BY clause)\n''', 
    335544826 : '''Nested aggregate and window functions are not allowed\n''', 
    335544827 : '''Invalid argument in EXECUTE STATEMENT - cannot convert to string\n''', 
    335544828 : '''Wrong request type in EXECUTE STATEMENT '@1'\n''', 
    335544829 : '''Variable type (position @1) in EXECUTE STATEMENT '@2' INTO does not match returned column type\n''', 
    335544830 : '''Too many recursion levels of EXECUTE STATEMENT\n''', 
    335544831 : '''Access to @1 "@2" is denied by server administrator\n''', 
    335544832 : '''Cannot change difference file name while database is in backup mode\n''', 
    335544833 : '''Physical backup is not allowed while Write-Ahead Log is in use\n''', 
    335544834 : '''Cursor is not open\n''', 
    335544835 : '''Target shutdown mode is invalid for database "@1"\n''', 
    335544836 : '''Concatenation overflow. Resulting string cannot exceed 32765 bytes in length.\n''', 
    335544837 : '''Invalid offset parameter @1 to SUBSTRING. Only positive integers are allowed.\n''', 
    335544838 : '''Foreign key reference target does not exist\n''', 
    335544839 : '''Foreign key references are present for the record\n''', 
    335544840 : '''cannot update\n''', 
    335544841 : '''Cursor is already open\n''', 
    335544842 : '''@1\n''', 
    335544843 : '''Context variable @1 is not found in namespace @2\n''', 
    335544844 : '''Invalid namespace name @1 passed to @2\n''', 
    335544845 : '''Too many context variables\n''', 
    335544846 : '''Invalid argument passed to @1\n''', 
    335544847 : '''BLR syntax error. Identifier @1... is too long\n''', 
    335544848 : '''exception @1\n''', 
    335544849 : '''Malformed string\n''', 
    335544850 : '''Output parameter mismatch for procedure @1\n''', 
    335544851 : '''Unexpected end of command - line @1, column @2\n''', 
    335544852 : '''partner index segment no @1 has incompatible data type\n''', 
    335544853 : '''Invalid length parameter @1 to SUBSTRING. Negative integers are not allowed.\n''', 
    335544854 : '''CHARACTER SET @1 is not installed\n''', 
    335544855 : '''COLLATION @1 for CHARACTER SET @2 is not installed\n''', 
    335544856 : '''connection shutdown\n''', 
    335544857 : '''Maximum BLOB size exceeded\n''', 
    335544858 : '''Can't have relation with only computed fields or constraints\n''', 
    335544859 : '''Time precision exceeds allowed range (0-@1)\n''', 
    335544860 : '''Unsupported conversion to target type BLOB (subtype @1)\n''', 
    335544861 : '''Unsupported conversion to target type ARRAY\n''', 
    335544862 : '''Stream does not support record locking\n''', 
    335544863 : '''Cannot create foreign key constraint @1. Partner index does not exist or is inactive.\n''', 
    335544864 : '''Transactions count exceeded. Perform backup and restore to make database operable again\n''', 
    335544865 : '''Column has been unexpectedly deleted\n''', 
    335544866 : '''@1 cannot depend on @2\n''', 
    335544867 : '''Blob sub_types bigger than 1 (text) are for internal use only\n''', 
    335544868 : '''Procedure @1 is not selectable (it does not contain a SUSPEND statement)\n''', 
    335544869 : '''Datatype @1 is not supported for sorting operation\n''', 
    335544870 : '''COLLATION @1\n''', 
    335544871 : '''DOMAIN @1\n''', 
    335544872 : '''domain @1 is not defined\n''', 
    335544873 : '''Array data type can use up to @1 dimensions\n''', 
    335544874 : '''A multi database transaction cannot span more than @1 databases\n''', 
    335544875 : '''Bad debug info format\n''', 
    335544876 : '''Error while parsing procedure @1's BLR\n''', 
    335544877 : '''index key too big\n''', 
    335544878 : '''concurrent transaction number is @1\n''', 
    335544879 : '''validation error for variable @1, value "@2"\n''', 
    335544880 : '''validation error for @1, value "@2"\n''', 
    335544881 : '''Difference file name should be set explicitly for database on raw device\n''', 
    335544882 : '''Login name too long (@1 characters, maximum allowed @2)\n''', 
    335544883 : '''column @1 is not defined in procedure @2\n''', 
    335544884 : '''Invalid SIMILAR TO pattern\n''', 
    335544885 : '''Invalid TEB format\n''', 
    335544886 : '''Found more than one transaction isolation in TPB\n''', 
    335544887 : '''Table reservation lock type @1 requires table name before in TPB\n''', 
    335544888 : '''Found more than one @1 specification in TPB\n''', 
    335544889 : '''Option @1 requires READ COMMITTED isolation in TPB\n''', 
    335544890 : '''Option @1 is not valid if @2 was used previously in TPB\n''', 
    335544891 : '''Table name length missing after table reservation @1 in TPB\n''', 
    335544892 : '''Table name length @1 is too long after table reservation @2 in TPB\n''', 
    335544893 : '''Table name length @1 without table name after table reservation @2 in TPB\n''', 
    335544894 : '''Table name length @1 goes beyond the remaining TPB size after table reservation @2\n''', 
    335544895 : '''Table name length is zero after table reservation @1 in TPB\n''', 
    335544896 : '''Table or view @1 not defined in system tables after table reservation @2 in TPB\n''', 
    335544897 : '''Base table or view @1 for view @2 not defined in system tables after table reservation @3 in TPB\n''', 
    335544898 : '''Option length missing after option @1 in TPB\n''', 
    335544899 : '''Option length @1 without value after option @2 in TPB\n''', 
    335544900 : '''Option length @1 goes beyond the remaining TPB size after option @2\n''', 
    335544901 : '''Option length is zero after table reservation @1 in TPB\n''', 
    335544902 : '''Option length @1 exceeds the range for option @2 in TPB\n''', 
    335544903 : '''Option value @1 is invalid for the option @2 in TPB\n''', 
    335544904 : '''Preserving previous table reservation @1 for table @2, stronger than new @3 in TPB\n''', 
    335544905 : '''Table reservation @1 for table @2 already specified and is stronger than new @3 in TPB\n''', 
    335544906 : '''Table reservation reached maximum recursion of @1 when expanding views in TPB\n''', 
    335544907 : '''Table reservation in TPB cannot be applied to @1 because it's a virtual table\n''', 
    335544908 : '''Table reservation in TPB cannot be applied to @1 because it's a system table\n''', 
    335544909 : '''Table reservation @1 or @2 in TPB cannot be applied to @3 because it's a temporary table\n''', 
    335544910 : '''Cannot set the transaction in read only mode after a table reservation isc_tpb_lock_write in TPB\n''', 
    335544911 : '''Cannot take a table reservation isc_tpb_lock_write in TPB because the transaction is in read only mode\n''', 
    335544912 : '''value exceeds the range for a valid time\n''', 
    335544913 : '''value exceeds the range for valid timestamps\n''', 
    335544914 : '''string right truncation\n''', 
    335544915 : '''blob truncation when converting to a string: length limit exceeded\n''', 
    335544916 : '''numeric value is out of range\n''', 
    335544917 : '''Firebird shutdown is still in progress after the specified timeout\n''', 
    335544918 : '''Attachment handle is busy\n''', 
    335544919 : '''Bad written UDF detected: pointer returned in FREE_IT function was not allocated by ib_util_malloc\n''', 
    335544920 : '''External Data Source provider '@1' not found\n''', 
    335544921 : '''Execute statement error at @1 :@2Data source : @3\n''', 
    335544922 : '''Execute statement preprocess SQL error\n''', 
    335544923 : '''Statement expected\n''', 
    335544924 : '''Parameter name expected\n''', 
    335544925 : '''Unclosed comment found near '@1'\n''', 
    335544926 : '''Execute statement error at @1 :@2Statement : @3Data source : @4\n''', 
    335544927 : '''Input parameters mismatch\n''', 
    335544928 : '''Output parameters mismatch\n''', 
    335544929 : '''Input parameter '@1' have no value set\n''', 
    335544930 : '''BLR stream length @1 exceeds implementation limit @2\n''', 
    335544931 : '''Monitoring table space exhausted\n''', 
    335544932 : '''module name or entrypoint could not be found\n''', 
    335544933 : '''nothing to cancel\n''', 
    335544934 : '''ib_util library has not been loaded to deallocate memory returned by FREE_IT function\n''', 
    335544935 : '''Cannot have circular dependencies with computed fields\n''', 
    335544936 : '''Security database error\n''', 
    335544937 : '''Invalid data type in DATE/TIME/TIMESTAMP addition or subtraction in add_datettime()\n''', 
    335544938 : '''Only a TIME value can be added to a DATE value\n''', 
    335544939 : '''Only a DATE value can be added to a TIME value\n''', 
    335544940 : '''TIMESTAMP values can be subtracted only from another TIMESTAMP value\n''', 
    335544941 : '''Only one operand can be of type TIMESTAMP\n''', 
    335544942 : '''Only HOUR, MINUTE, SECOND and MILLISECOND can be extracted from TIME values\n''', 
    335544943 : '''HOUR, MINUTE, SECOND and MILLISECOND cannot be extracted from DATE values\n''', 
    335544944 : '''Invalid argument for EXTRACT() not being of DATE/TIME/TIMESTAMP type\n''', 
    335544945 : '''Arguments for @1 must be integral types or NUMERIC/DECIMAL without scale\n''', 
    335544946 : '''First argument for @1 must be integral type or floating point type\n''', 
    335544947 : '''Human readable UUID argument for @1 must be of string type\n''', 
    335544948 : '''Human readable UUID argument for @2 must be of exact length @1\n''', 
    335544949 : '''Human readable UUID argument for @3 must have "-" at position @2 instead of "@1"\n''', 
    335544950 : '''Human readable UUID argument for @3 must have hex digit at position @2 instead of "@1"\n''', 
    335544951 : '''Only HOUR, MINUTE, SECOND and MILLISECOND can be added to TIME values in @1\n''', 
    335544952 : '''Invalid data type in addition of part to DATE/TIME/TIMESTAMP in @1\n''', 
    335544953 : '''Invalid part @1 to be added to a DATE/TIME/TIMESTAMP value in @2\n''', 
    335544954 : '''Expected DATE/TIME/TIMESTAMP type in evlDateAdd() result\n''', 
    335544955 : '''Expected DATE/TIME/TIMESTAMP type as first and second argument to @1\n''', 
    335544956 : '''The result of TIME-<value> in @1 cannot be expressed in YEAR, MONTH, DAY or WEEK\n''', 
    335544957 : '''The result of TIME-TIMESTAMP or TIMESTAMP-TIME in @1 cannot be expressed in HOUR, MINUTE, SECOND or MILLISECOND\n''', 
    335544958 : '''The result of DATE-TIME or TIME-DATE in @1 cannot be expressed in HOUR, MINUTE, SECOND and MILLISECOND\n''', 
    335544959 : '''Invalid part @1 to express the difference between two DATE/TIME/TIMESTAMP values in @2\n''', 
    335544960 : '''Argument for @1 must be positive\n''', 
    335544961 : '''Base for @1 must be positive\n''', 
    335544962 : '''Argument #@1 for @2 must be zero or positive\n''', 
    335544963 : '''Argument #@1 for @2 must be positive\n''', 
    335544964 : '''Base for @1 cannot be zero if exponent is negative\n''', 
    335544965 : '''Base for @1 cannot be negative if exponent is not an integral value\n''', 
    335544966 : '''The numeric scale must be between -128 and 127 in @1\n''', 
    335544967 : '''Argument for @1 must be zero or positive\n''', 
    335544968 : '''Binary UUID argument for @1 must be of string type\n''', 
    335544969 : '''Binary UUID argument for @2 must use @1 bytes\n''', 
    335544970 : '''Missing required item @1 in service parameter block\n''', 
    335544971 : '''@1 server is shutdown\n''', 
    335544972 : '''Invalid connection string\n''', 
    335544973 : '''Unrecognized events block\n''', 
    335544974 : '''Could not start first worker thread - shutdown server\n''', 
    335544975 : '''Timeout occurred while waiting for a secondary connection for event processing\n''', 
    335544976 : '''Argument for @1 must be different than zero\n''', 
    335544977 : '''Argument for @1 must be in the range [-1, 1]\n''', 
    335544978 : '''Argument for @1 must be greater or equal than one\n''', 
    335544979 : '''Argument for @1 must be in the range ]-1, 1[\n''', 
    335544980 : '''Incorrect parameters provided to internal function @1\n''', 
    335544981 : '''Floating point overflow in built-in function @1\n''', 
    335544982 : '''Floating point overflow in result from UDF @1\n''', 
    335544983 : '''Invalid floating point value returned by UDF @1\n''', 
    335544984 : '''Database is probably already opened by another engine instance in another Windows session\n''', 
    335544985 : '''No free space found in temporary directories\n''', 
    335544986 : '''Explicit transaction control is not allowed\n''', 
    335544987 : '''Use of TRUSTED switches in spb_command_line is prohibited\n''', 
    335544988 : '''PACKAGE @1\n''', 
    335544989 : '''Cannot make field @1 of table @2 NOT NULL because there are NULLs present\n''', 
    335544990 : '''Feature @1 is not supported anymore\n''', 
    335544991 : '''VIEW @1\n''', 
    335544992 : '''Can not access lock files directory @1\n''', 
    335544993 : '''Fetch option @1 is invalid for a non-scrollable cursor\n''', 
    335544994 : '''Error while parsing function @1's BLR\n''', 
    335544995 : '''Cannot execute function @1 of the unimplemented package @2\n''', 
    335544996 : '''Cannot execute procedure @1 of the unimplemented package @2\n''', 
    335544997 : '''External function @1 not returned by the external engine plugin @2\n''', 
    335544998 : '''External procedure @1 not returned by the external engine plugin @2\n''', 
    335544999 : '''External trigger @1 not returned by the external engine plugin @2\n''', 
    335545000 : '''Incompatible plugin version @1 for external engine @2\n''', 
    335545001 : '''External engine @1 not found\n''', 
    335545002 : '''Attachment is in use\n''', 
    335545003 : '''Transaction is in use\n''', 
    335545004 : '''Plugin @1 not found\n''', 
    335545005 : '''Module @1 exists, but can not be loaded\n''', 
    335545006 : '''Entrypoint of plugin @1 does not exist\n''', 
    335545007 : '''Invalid value @1 for parameter index at PluginImpl::getConfigInfo: out of bounds\n''', 
    335545008 : '''Plugin @1 does not create @2 instances\n''', 
    335545009 : '''Invalid usage of context namespace DDL_TRIGGER\n''', 
    335545010 : '''Value is NULL but isNull parameter was not informed\n''', 
    335545011 : '''Type @1 is incompatible with BLOB\n''', 
    335545012 : '''Invalid date\n''', 
    335545013 : '''Invalid time\n''', 
    335545014 : '''Invalid timestamp\n''', 
    335545015 : '''Invalid index @1 in function @2\n''', 
    335545016 : '''@1\n''', 
    335545017 : '''Asynchronous call is already running for this attachment\n''', 
    335545018 : '''Function @1 is private to package @2\n''', 
    335545019 : '''Procedure @1 is private to package @2\n''', 
    335545020 : '''Request can't access new records in relation @1 and should be recompiled\n''', 
    335545021 : '''invalid events id (handle)\n''', 
    335545022 : '''Cannot copy statement @1\n''', 
    335545023 : '''Invalid usage of boolean expression\n''', 
    335545024 : '''Arguments for @1 cannot both be zero\n''', 
    335545025 : '''missing service ID in spb\n''', 
    335545026 : '''External BLR message mismatch: invalid null descriptor at field @1\n''', 
    335545027 : '''External BLR message mismatch: length = @1, expected @2\n''', 
    335545028 : '''Subscript @1 out of bounds [@2, @3]\n''', 
    335545029 : '''Install incomplete, please read chapter "Initializing security database" in Quick Start Guide\n''', 
    335545030 : '''@1 operation is not allowed for system table @2\n''', 
    335545031 : '''Libtommath error code @1 in function @2\n''', 
    335545032 : '''unsupported BLR version (expected between @1 and @2, encountered @3)\n''', 
    335545033 : '''expected length @1, actual @2\n''', 
    335545034 : '''Wrong info requested in isc_svc_query() for anonymous service\n''', 
    335545035 : '''No isc_info_svc_stdin in user request, but service thread requested stdin data\n''', 
    335545036 : '''Start request for anonymous service is impossible\n''', 
    335545037 : '''All services except for getting server log require switches\n''', 
    335545038 : '''Size of stdin data is more than was requested from client\n''', 
    335545039 : '''Crypt plugin @1 failed to load\n''', 
    335545040 : '''Length of crypt plugin name should not exceed @1 bytes\n''', 
    335545041 : '''Crypt failed - already crypting database\n''', 
    335545042 : '''Crypt failed - database is already in requested state\n''', 
    335545043 : '''Missing crypt plugin, but page appears encrypted\n''', 
    335545044 : '''No providers loaded\n''', 
    335545045 : '''NULL data with non-zero SPB length\n''', 
    335545046 : '''Maximum (@1) number of arguments exceeded for function @2\n''', 
    335545047 : '''External BLR message mismatch: names count = @1, blr count = @2\n''', 
    335545048 : '''External BLR message mismatch: name @1 not found\n''', 
    335545049 : '''Invalid resultset interface\n''', 
    335545050 : '''Message length passed from user application does not match set of columns\n''', 
    335545051 : '''Resultset is missing output format information\n''', 
    335545052 : '''Message metadata not ready - item @1 is not finished\n''', 
    335545053 : '''Missing configuration file: @1\n''', 
    335545054 : '''@1: illegal line <@2>\n''', 
    335545055 : '''Invalid include operator in @1 for <@2>\n''', 
    335545056 : '''Include depth too big\n''', 
    335545057 : '''File to include not found\n''', 
    335545058 : '''Only the owner can change the ownership\n''', 
    335545059 : '''undefined variable number\n''', 
    335545060 : '''Missing security context for database @1\n''', 
    335545061 : '''Missing segment @1 in multisegment connect block parameter\n''', 
    335545062 : '''Different logins in connect and attach packets - client library error\n''', 
    335545063 : '''Exceeded exchange limit during authentication handshake\n''', 
    335545064 : '''Incompatible wire encryption levels requested on client and server\n''', 
    335545065 : '''Client attempted to attach unencrypted but wire encryption is required\n''', 
    335545066 : '''Client attempted to start wire encryption using unknown key @1\n''', 
    335545067 : '''Client attempted to start wire encryption using unsupported plugin @1\n''', 
    335545068 : '''Error getting security database name from configuration file\n''', 
    335545069 : '''Client authentication plugin is missing required data from server\n''', 
    335545070 : '''Client authentication plugin expected @2 bytes of @3 from server, got @1\n''', 
    335545071 : '''Attempt to get information about an unprepared dynamic SQL statement.\n''', 
    335545072 : '''Problematic key value is @1\n''', 
    335740929 : '''data base file name (@1) already given\n''', 
    335740930 : '''invalid switch @1\n''', 
    335740932 : '''incompatible switch combination\n''', 
    335740933 : '''replay log pathname required\n''', 
    335740934 : '''number of page buffers for cache required\n''', 
    335740935 : '''numeric value required\n''', 
    335740936 : '''positive numeric value required\n''', 
    335740937 : '''number of transactions per sweep required\n''', 
    335740940 : '''"full" or "reserve" required\n''', 
    335740941 : '''user name required\n''', 
    335740942 : '''password required\n''', 
    335740943 : '''subsystem name\n''', 
    335740944 : '''"wal" required\n''', 
    335740945 : '''number of seconds required\n''', 
    335740946 : '''numeric value between 0 and 32767 inclusive required\n''', 
    335740947 : '''must specify type of shutdown\n''', 
    335740948 : '''please retry, specifying an option\n''', 
    335740951 : '''please retry, giving a database name\n''', 
    335740991 : '''internal block exceeds maximum size\n''', 
    335740992 : '''corrupt pool\n''', 
    335740993 : '''virtual memory exhausted\n''', 
    335740994 : '''bad pool id\n''', 
    335740995 : '''Transaction state @1 not in valid range.\n''', 
    335741012 : '''unexpected end of input\n''', 
    335741018 : '''failed to reconnect to a transaction in database @1\n''', 
    335741036 : '''Transaction description item unknown\n''', 
    335741038 : '''"read_only" or "read_write" required\n''', 
    335741042 : '''positive or zero numeric value required\n''', 
    336003074 : '''Cannot SELECT RDB$DB_KEY from a stored procedure.\n''', 
    336003075 : '''Precision 10 to 18 changed from DOUBLE PRECISION in SQL dialect 1 to 64-bit scaled integer in SQL dialect 3\n''', 
    336003076 : '''Use of @1 expression that returns different results in dialect 1 and dialect 3\n''', 
    336003077 : '''Database SQL dialect @1 does not support reference to @2 datatype\n''', 
    336003079 : '''DB dialect @1 and client dialect @2 conflict with respect to numeric precision @3.\n''', 
    336003080 : '''WARNING: Numeric literal @1 is interpreted as a floating-point\n''', 
    336003081 : '''value in SQL dialect 1, but as an exact numeric value in SQL dialect 3.\n''', 
    336003082 : '''WARNING: NUMERIC and DECIMAL fields with precision 10 or greater are stored\n''', 
    336003083 : '''as approximate floating-point values in SQL dialect 1, but as 64-bit\n''', 
    336003084 : '''integers in SQL dialect 3.\n''', 
    336003085 : '''Ambiguous field name between @1 and @2\n''', 
    336003086 : '''External function should have return position between 1 and @1\n''', 
    336003087 : '''Label @1 @2 in the current scope\n''', 
    336003088 : '''Datatypes @1are not comparable in expression @2\n''', 
    336003089 : '''Empty cursor name is not allowed\n''', 
    336003090 : '''Statement already has a cursor @1 assigned\n''', 
    336003091 : '''Cursor @1 is not found in the current context\n''', 
    336003092 : '''Cursor @1 already exists in the current context\n''', 
    336003093 : '''Relation @1 is ambiguous in cursor @2\n''', 
    336003094 : '''Relation @1 is not found in cursor @2\n''', 
    336003095 : '''Cursor is not open\n''', 
    336003096 : '''Data type @1 is not supported for EXTERNAL TABLES. Relation '@2', field '@3'\n''', 
    336003097 : '''Feature not supported on ODS version older than @1.@2\n''', 
    336003098 : '''Primary key required on table @1\n''', 
    336003099 : '''UPDATE OR INSERT field list does not match primary key of table @1\n''', 
    336003100 : '''UPDATE OR INSERT field list does not match MATCHING clause\n''', 
    336003101 : '''UPDATE OR INSERT without MATCHING could not be used with views based on more than one table\n''', 
    336003102 : '''Incompatible trigger type\n''', 
    336003103 : '''Database trigger type can't be changed\n''', 
    336003104 : '''To be used with RDB$RECORD_VERSION, @1 must be a table or a view of single table\n''', 
    336068645 : '''BLOB Filter @1 not found\n''', 
    336068649 : '''Function @1 not found\n''', 
    336068656 : '''Index not found\n''', 
    336068662 : '''View @1 not found\n''', 
    336068697 : '''Domain not found\n''', 
    336068717 : '''Triggers created automatically cannot be modified\n''', 
    336068740 : '''Table @1 already exists\n''', 
    336068748 : '''Procedure @1 not found\n''', 
    336068752 : '''Exception not found\n''', 
    336068754 : '''Parameter @1 in procedure @2 not found\n''', 
    336068755 : '''Trigger @1 not found\n''', 
    336068759 : '''Character set @1 not found\n''', 
    336068760 : '''Collation @1 not found\n''', 
    336068763 : '''Role @1 not found\n''', 
    336068767 : '''Name longer than database column size\n''', 
    336068784 : '''column @1 does not exist in table/view @2\n''', 
    336068796 : '''SQL role @1 does not exist\n''', 
    336068797 : '''user @1 has no grant admin option on SQL role @2\n''', 
    336068798 : '''user @1 is not a member of SQL role @2\n''', 
    336068799 : '''@1 is not the owner of SQL role @2\n''', 
    336068800 : '''@1 is a SQL role and not a user\n''', 
    336068801 : '''user name @1 could not be used for SQL role\n''', 
    336068802 : '''SQL role @1 already exists\n''', 
    336068803 : '''keyword @1 can not be used as a SQL role name\n''', 
    336068804 : '''SQL roles are not supported in on older versions of the database.  A backup and restore of the database is required.\n''', 
    336068812 : '''Cannot rename domain @1 to @2.  A domain with that name already exists.\n''', 
    336068813 : '''Cannot rename column @1 to @2.  A column with that name already exists in table @3.\n''', 
    336068814 : '''Column @1 from table @2 is referenced in @3\n''', 
    336068815 : '''Cannot change datatype for column @1.  Changing datatype is not supported for BLOB or ARRAY columns.\n''', 
    336068816 : '''New size specified for column @1 must be at least @2 characters.\n''', 
    336068817 : '''Cannot change datatype for @1.  Conversion from base type @2 to @3 is not supported.\n''', 
    336068818 : '''Cannot change datatype for column @1 from a character type to a non-character type.\n''', 
    336068820 : '''Zero length identifiers are not allowed\n''', 
    336068822 : '''Sequence @1 not found\n''', 
    336068829 : '''Maximum number of collations per character set exceeded\n''', 
    336068830 : '''Invalid collation attributes\n''', 
    336068840 : '''@1 cannot reference @2\n''', 
    336068843 : '''Collation @1 is used in table @2 (field name @3) and cannot be dropped\n''', 
    336068844 : '''Collation @1 is used in domain @2 and cannot be dropped\n''', 
    336068845 : '''Cannot delete system collation\n''', 
    336068846 : '''Cannot delete default collation of CHARACTER SET @1\n''', 
    336068849 : '''Table @1 not found\n''', 
    336068851 : '''Collation @1 is used in procedure @2 (parameter name @3) and cannot be dropped\n''', 
    336068852 : '''New scale specified for column @1 must be at most @2.\n''', 
    336068853 : '''New precision specified for column @1 must be at least @2.\n''', 
    336068855 : '''Warning: @1 on @2 is not granted to @3.\n''', 
    336068856 : '''Feature '@1' is not supported in ODS @2.@3\n''', 
    336068857 : '''Cannot add or remove COMPUTED from column @1\n''', 
    336068858 : '''Password should not be empty string\n''', 
    336068859 : '''Index @1 already exists\n''', 
    336068864 : '''Package @1 not found\n''', 
    336068865 : '''Schema @1 not found\n''', 
    336068866 : '''Cannot ALTER or DROP system procedure @1\n''', 
    336068867 : '''Cannot ALTER or DROP system trigger @1\n''', 
    336068868 : '''Cannot ALTER or DROP system function @1\n''', 
    336068869 : '''Invalid DDL statement for procedure @1\n''', 
    336068870 : '''Invalid DDL statement for trigger @1\n''', 
    336068871 : '''Function @1 has not been defined on the package body @2\n''', 
    336068872 : '''Procedure @1 has not been defined on the package body @2\n''', 
    336068873 : '''Function @1 has a signature mismatch on package body @2\n''', 
    336068874 : '''Procedure @1 has a signature mismatch on package body @2\n''', 
    336068875 : '''Default values for parameters are allowed only in declaration of packaged procedure @1.@2\n''', 
    336068877 : '''Package body @1 already exists\n''', 
    336068878 : '''Invalid DDL statement for function @1\n''', 
    336068879 : '''Cannot alter new style function @1 with ALTER EXTERNAL FUNCTION. Use ALTER FUNCTION instead.\n''', 
    336068886 : '''Parameter @1 in function @2 not found\n''', 
    336068887 : '''Parameter @1 of routine @2 not found\n''', 
    336068888 : '''Parameter @1 of routine @2 is ambiguous (found in both procedures and functions). Use a specifier keyword.\n''', 
    336068889 : '''Collation @1 is used in function @2 (parameter name @3) and cannot be dropped\n''', 
    336068890 : '''Domain @1 is used in function @2 (parameter name @3) and cannot be dropped\n''', 
    336068891 : '''ALTER USER requires at least one clause to be specified\n''', 
    336330753 : '''found unknown switch\n''', 
    336330754 : '''page size parameter missing\n''', 
    336330755 : '''Page size specified (@1) greater than limit (16384 bytes)\n''', 
    336330756 : '''redirect location for output is not specified\n''', 
    336330757 : '''conflicting switches for backup/restore\n''', 
    336330758 : '''device type @1 not known\n''', 
    336330759 : '''protection is not there yet\n''', 
    336330760 : '''page size is allowed only on restore or create\n''', 
    336330761 : '''multiple sources or destinations specified\n''', 
    336330762 : '''requires both input and output filenames\n''', 
    336330763 : '''input and output have the same name.  Disallowed.\n''', 
    336330764 : '''expected page size, encountered "@1"\n''', 
    336330765 : '''REPLACE specified, but the first file @1 is a database\n''', 
    336330766 : '''database @1 already exists.  To replace it, use the -REP switch\n''', 
    336330767 : '''device type not specified\n''', 
    336330772 : '''gds_$blob_info failed\n''', 
    336330773 : '''do not understand BLOB INFO item @1\n''', 
    336330774 : '''gds_$get_segment failed\n''', 
    336330775 : '''gds_$close_blob failed\n''', 
    336330776 : '''gds_$open_blob failed\n''', 
    336330777 : '''Failed in put_blr_gen_id\n''', 
    336330778 : '''data type @1 not understood\n''', 
    336330779 : '''gds_$compile_request failed\n''', 
    336330780 : '''gds_$start_request failed\n''', 
    336330781 : '''gds_$receive failed\n''', 
    336330782 : '''gds_$release_request failed\n''', 
    336330783 : '''gds_$database_info failed\n''', 
    336330784 : '''Expected database description record\n''', 
    336330785 : '''failed to create database @1\n''', 
    336330786 : '''RESTORE: decompression length error\n''', 
    336330787 : '''cannot find table @1\n''', 
    336330788 : '''Cannot find column for BLOB\n''', 
    336330789 : '''gds_$create_blob failed\n''', 
    336330790 : '''gds_$put_segment failed\n''', 
    336330791 : '''expected record length\n''', 
    336330792 : '''wrong length record, expected @1 encountered @2\n''', 
    336330793 : '''expected data attribute\n''', 
    336330794 : '''Failed in store_blr_gen_id\n''', 
    336330795 : '''do not recognize record type @1\n''', 
    336330796 : '''Expected backup version 1..10.  Found @1\n''', 
    336330797 : '''expected backup description record\n''', 
    336330798 : '''string truncated\n''', 
    336330799 : '''warning -- record could not be restored\n''', 
    336330800 : '''gds_$send failed\n''', 
    336330801 : '''no table name for data\n''', 
    336330802 : '''unexpected end of file on backup file\n''', 
    336330803 : '''database format @1 is too old to restore to\n''', 
    336330804 : '''array dimension for column @1 is invalid\n''', 
    336330807 : '''Expected XDR record length\n''', 
    336330817 : '''cannot open backup file @1\n''', 
    336330818 : '''cannot open status and error output file @1\n''', 
    336330934 : '''blocking factor parameter missing\n''', 
    336330935 : '''expected blocking factor, encountered "@1"\n''', 
    336330936 : '''a blocking factor may not be used in conjunction with device CT\n''', 
    336330940 : '''user name parameter missing\n''', 
    336330941 : '''password parameter missing\n''', 
    336330952 : ''' missing parameter for the number of bytes to be skipped\n''', 
    336330953 : '''expected number of bytes to be skipped, encountered "@1"\n''', 
    336330965 : '''character set\n''', 
    336330967 : '''collation\n''', 
    336330972 : '''Unexpected I/O error while reading from backup file\n''', 
    336330973 : '''Unexpected I/O error while writing to backup file\n''', 
    336330985 : '''could not drop database @1 (database might be in use)\n''', 
    336330990 : '''System memory exhausted\n''', 
    336331002 : '''SQL role\n''', 
    336331005 : '''SQL role parameter missing\n''', 
    336331010 : '''page buffers parameter missing\n''', 
    336331011 : '''expected page buffers, encountered "@1"\n''', 
    336331012 : '''page buffers is allowed only on restore or create\n''', 
    336331014 : '''size specification either missing or incorrect for file @1\n''', 
    336331015 : '''file @1 out of sequence\n''', 
    336331016 : '''can't join -- one of the files missing\n''', 
    336331017 : ''' standard input is not supported when using join operation\n''', 
    336331018 : '''standard output is not supported when using split operation or in verbose mode\n''', 
    336331019 : '''backup file @1 might be corrupt\n''', 
    336331020 : '''database file specification missing\n''', 
    336331021 : '''can't write a header record to file @1\n''', 
    336331022 : '''free disk space exhausted\n''', 
    336331023 : '''file size given (@1) is less than minimum allowed (@2)\n''', 
    336331025 : '''service name parameter missing\n''', 
    336331026 : '''Cannot restore over current database, must be SYSDBA or owner of the existing database.\n''', 
    336331031 : '''"read_only" or "read_write" required\n''', 
    336331033 : '''just data ignore all constraints etc.\n''', 
    336331034 : '''restoring data only ignoring foreign key, unique, not null & other constraints\n''', 
    336331078 : '''verbose interval value parameter missing\n''', 
    336331079 : '''verbose interval value cannot be smaller than @1\n''', 
    336331081 : '''verify (verbose) and verbint options are mutually exclusive\n''', 
    336331082 : '''option -@1 is allowed only on restore or create\n''', 
    336331083 : '''option -@1 is allowed only on backup\n''', 
    336331084 : '''options -@1 and -@2 are mutually exclusive\n''', 
    336331085 : '''parameter for option -@1 was already specified with value "@2"\n''', 
    336331086 : '''option -@1 was already specified\n''', 
    336331091 : '''dependency depth greater than @1 for view @2\n''', 
    336331092 : '''value greater than @1 when calculating length of rdb$db_key for view @2\n''', 
    336331093 : '''Invalid metadata detected. Use -FIX_FSS_METADATA option.\n''', 
    336331094 : '''Invalid data detected. Use -FIX_FSS_DATA option.\n''', 
    336331096 : '''Expected backup version @2..@3.  Found @1\n''', 
    336331100 : '''database format @1 is too old to backup\n''', 
    336397205 : '''ODS versions before ODS@1 are not supported\n''', 
    336397206 : '''Table @1 does not exist\n''', 
    336397207 : '''View @1 does not exist\n''', 
    336397208 : '''At line @1, column @2\n''', 
    336397209 : '''At unknown line and column\n''', 
    336397210 : '''Column @1 cannot be repeated in @2 statement\n''', 
    336397211 : '''Too many values (more than @1) in member list to match against\n''', 
    336397212 : '''Array and BLOB data types not allowed in computed field\n''', 
    336397213 : '''Implicit domain name @1 not allowed in user created domain\n''', 
    336397214 : '''scalar operator used on field @1 which is not an array\n''', 
    336397215 : '''cannot sort on more than 255 items\n''', 
    336397216 : '''cannot group on more than 255 items\n''', 
    336397217 : '''Cannot include the same field (@1.@2) twice in the ORDER BY clause with conflicting sorting options\n''', 
    336397218 : '''column list from derived table @1 has more columns than the number of items in its SELECT statement\n''', 
    336397219 : '''column list from derived table @1 has less columns than the number of items in its SELECT statement\n''', 
    336397220 : '''no column name specified for column number @1 in derived table @2\n''', 
    336397221 : '''column @1 was specified multiple times for derived table @2\n''', 
    336397222 : '''Internal dsql error: alias type expected by pass1_expand_select_node\n''', 
    336397223 : '''Internal dsql error: alias type expected by pass1_field\n''', 
    336397224 : '''Internal dsql error: column position out of range in pass1_union_auto_cast\n''', 
    336397225 : '''Recursive CTE member (@1) can refer itself only in FROM clause\n''', 
    336397226 : '''CTE '@1' has cyclic dependencies\n''', 
    336397227 : '''Recursive member of CTE can't be member of an outer join\n''', 
    336397228 : '''Recursive member of CTE can't reference itself more than once\n''', 
    336397229 : '''Recursive CTE (@1) must be an UNION\n''', 
    336397230 : '''CTE '@1' defined non-recursive member after recursive\n''', 
    336397231 : '''Recursive member of CTE '@1' has @2 clause\n''', 
    336397232 : '''Recursive members of CTE (@1) must be linked with another members via UNION ALL\n''', 
    336397233 : '''Non-recursive member is missing in CTE '@1'\n''', 
    336397234 : '''WITH clause can't be nested\n''', 
    336397235 : '''column @1 appears more than once in USING clause\n''', 
    336397236 : '''feature is not supported in dialect @1\n''', 
    336397237 : '''CTE "@1" is not used in query\n''', 
    336397238 : '''column @1 appears more than once in ALTER VIEW\n''', 
    336397239 : '''@1 is not supported inside IN AUTONOMOUS TRANSACTION block\n''', 
    336397240 : '''Unknown node type @1 in dsql/GEN_expr\n''', 
    336397241 : '''Argument for @1 in dialect 1 must be string or numeric\n''', 
    336397242 : '''Argument for @1 in dialect 3 must be numeric\n''', 
    336397243 : '''Strings cannot be added to or subtracted from DATE or TIME types\n''', 
    336397244 : '''Invalid data type for subtraction involving DATE, TIME or TIMESTAMP types\n''', 
    336397245 : '''Adding two DATE values or two TIME values is not allowed\n''', 
    336397246 : '''DATE value cannot be subtracted from the provided data type\n''', 
    336397247 : '''Strings cannot be added or subtracted in dialect 3\n''', 
    336397248 : '''Invalid data type for addition or subtraction in dialect 3\n''', 
    336397249 : '''Invalid data type for multiplication in dialect 1\n''', 
    336397250 : '''Strings cannot be multiplied in dialect 3\n''', 
    336397251 : '''Invalid data type for multiplication in dialect 3\n''', 
    336397252 : '''Division in dialect 1 must be between numeric data types\n''', 
    336397253 : '''Strings cannot be divided in dialect 3\n''', 
    336397254 : '''Invalid data type for division in dialect 3\n''', 
    336397255 : '''Strings cannot be negated (applied the minus operator) in dialect 3\n''', 
    336397256 : '''Invalid data type for negation (minus operator)\n''', 
    336397257 : '''Cannot have more than 255 items in DISTINCT list\n''', 
    336397258 : '''ALTER CHARACTER SET @1 failed\n''', 
    336397259 : '''COMMENT ON @1 failed\n''', 
    336397260 : '''CREATE FUNCTION @1 failed\n''', 
    336397261 : '''ALTER FUNCTION @1 failed\n''', 
    336397262 : '''CREATE OR ALTER FUNCTION @1 failed\n''', 
    336397263 : '''DROP FUNCTION @1 failed\n''', 
    336397264 : '''RECREATE FUNCTION @1 failed\n''', 
    336397265 : '''CREATE PROCEDURE @1 failed\n''', 
    336397266 : '''ALTER PROCEDURE @1 failed\n''', 
    336397267 : '''CREATE OR ALTER PROCEDURE @1 failed\n''', 
    336397268 : '''DROP PROCEDURE @1 failed\n''', 
    336397269 : '''RECREATE PROCEDURE @1 failed\n''', 
    336397270 : '''CREATE TRIGGER @1 failed\n''', 
    336397271 : '''ALTER TRIGGER @1 failed\n''', 
    336397272 : '''CREATE OR ALTER TRIGGER @1 failed\n''', 
    336397273 : '''DROP TRIGGER @1 failed\n''', 
    336397274 : '''RECREATE TRIGGER @1 failed\n''', 
    336397275 : '''CREATE COLLATION @1 failed\n''', 
    336397276 : '''DROP COLLATION @1 failed\n''', 
    336397277 : '''CREATE DOMAIN @1 failed\n''', 
    336397278 : '''ALTER DOMAIN @1 failed\n''', 
    336397279 : '''DROP DOMAIN @1 failed\n''', 
    336397280 : '''CREATE EXCEPTION @1 failed\n''', 
    336397281 : '''ALTER EXCEPTION @1 failed\n''', 
    336397282 : '''CREATE OR ALTER EXCEPTION @1 failed\n''', 
    336397283 : '''RECREATE EXCEPTION @1 failed\n''', 
    336397284 : '''DROP EXCEPTION @1 failed\n''', 
    336397285 : '''CREATE SEQUENCE @1 failed\n''', 
    336397286 : '''CREATE TABLE @1 failed\n''', 
    336397287 : '''ALTER TABLE @1 failed\n''', 
    336397288 : '''DROP TABLE @1 failed\n''', 
    336397289 : '''RECREATE TABLE @1 failed\n''', 
    336397290 : '''CREATE PACKAGE @1 failed\n''', 
    336397291 : '''ALTER PACKAGE @1 failed\n''', 
    336397292 : '''CREATE OR ALTER PACKAGE @1 failed\n''', 
    336397293 : '''DROP PACKAGE @1 failed\n''', 
    336397294 : '''RECREATE PACKAGE @1 failed\n''', 
    336397295 : '''CREATE PACKAGE BODY @1 failed\n''', 
    336397296 : '''DROP PACKAGE BODY @1 failed\n''', 
    336397297 : '''RECREATE PACKAGE BODY @1 failed\n''', 
    336397298 : '''CREATE VIEW @1 failed\n''', 
    336397299 : '''ALTER VIEW @1 failed\n''', 
    336397300 : '''CREATE OR ALTER VIEW @1 failed\n''', 
    336397301 : '''RECREATE VIEW @1 failed\n''', 
    336397302 : '''DROP VIEW @1 failed\n''', 
    336397303 : '''DROP SEQUENCE @1 failed\n''', 
    336397304 : '''RECREATE SEQUENCE @1 failed\n''', 
    336397305 : '''DROP INDEX @1 failed\n''', 
    336397306 : '''DROP FILTER @1 failed\n''', 
    336397307 : '''DROP SHADOW @1 failed\n''', 
    336397308 : '''DROP ROLE @1 failed\n''', 
    336397309 : '''DROP USER @1 failed\n''', 
    336397310 : '''CREATE ROLE @1 failed\n''', 
    336397311 : '''ALTER ROLE @1 failed\n''', 
    336397312 : '''ALTER INDEX @1 failed\n''', 
    336397313 : '''ALTER DATABASE failed\n''', 
    336397314 : '''CREATE SHADOW @1 failed\n''', 
    336397315 : '''DECLARE FILTER @1 failed\n''', 
    336397316 : '''CREATE INDEX @1 failed\n''', 
    336397317 : '''CREATE USER @1 failed\n''', 
    336397318 : '''ALTER USER @1 failed\n''', 
    336397319 : '''GRANT failed\n''', 
    336397320 : '''REVOKE failed\n''', 
    336397321 : '''Recursive member of CTE cannot use aggregate or window function\n''', 
    336723983 : '''unable to open database\n''', 
    336723984 : '''error in switch specifications\n''', 
    336723985 : '''no operation specified\n''', 
    336723986 : '''no user name specified\n''', 
    336723987 : '''add record error\n''', 
    336723988 : '''modify record error\n''', 
    336723989 : '''find/modify record error\n''', 
    336723990 : '''record not found for user: @1\n''', 
    336723991 : '''delete record error\n''', 
    336723992 : '''find/delete record error\n''', 
    336723996 : '''find/display record error\n''', 
    336723997 : '''invalid parameter, no switch defined\n''', 
    336723998 : '''operation already specified\n''', 
    336723999 : '''password already specified\n''', 
    336724000 : '''uid already specified\n''', 
    336724001 : '''gid already specified\n''', 
    336724002 : '''project already specified\n''', 
    336724003 : '''organization already specified\n''', 
    336724004 : '''first name already specified\n''', 
    336724005 : '''middle name already specified\n''', 
    336724006 : '''last name already specified\n''', 
    336724008 : '''invalid switch specified\n''', 
    336724009 : '''ambiguous switch specified\n''', 
    336724010 : '''no operation specified for parameters\n''', 
    336724011 : '''no parameters allowed for this operation\n''', 
    336724012 : '''incompatible switches specified\n''', 
    336724044 : '''Invalid user name (maximum 31 bytes allowed)\n''', 
    336724045 : '''Warning - maximum 8 significant bytes of password used\n''', 
    336724046 : '''database already specified\n''', 
    336724047 : '''database administrator name already specified\n''', 
    336724048 : '''database administrator password already specified\n''', 
    336724049 : '''SQL role name already specified\n''', 
    336920577 : '''found unknown switch\n''', 
    336920578 : '''please retry, giving a database name\n''', 
    336920579 : '''Wrong ODS version, expected @1, encountered @2\n''', 
    336920580 : '''Unexpected end of database file.\n''', 
    336920605 : '''Can't open database file @1\n''', 
    336920606 : '''Can't read a database page\n''', 
    336920607 : '''System memory exhausted\n''', 
    336986113 : '''Wrong value for access mode\n''', 
    336986114 : '''Wrong value for write mode\n''', 
    336986115 : '''Wrong value for reserve space\n''', 
    336986116 : '''Unknown tag (@1) in info_svr_db_info block after isc_svc_query()\n''', 
    336986117 : '''Unknown tag (@1) in isc_svc_query() results\n''', 
    336986118 : '''Unknown switch "@1"\n''', 
    336986159 : '''Wrong value for shutdown mode\n''', 
    336986160 : '''could not open file @1\n''', 
    336986161 : '''could not read file @1\n''', 
    336986162 : '''empty file @1\n''', 
    336986164 : '''Invalid or missing parameter for switch @1\n''', 
    337051649 : '''Switches trusted_user and trusted_role are not supported from command line\n''', 
    337117213 : '''Missing parameter for switch @1\n''', 
    337117214 : '''Only one of -LOCK, -UNLOCK, -FIXUP, -BACKUP or -RESTORE should be specified\n''', 
    337117215 : '''Unrecognized parameter @1\n''', 
    337117216 : '''Unknown switch @1\n''', 
    337117217 : '''Fetch password can't be used in service mode\n''', 
    337117218 : '''Error working with password file "@1"\n''', 
    337117219 : '''Switch -SIZE can be used only with -LOCK\n''', 
    337117220 : '''None of -LOCK, -UNLOCK, -FIXUP, -BACKUP or -RESTORE specified\n''', 
    337117223 : '''IO error reading file: @1\n''', 
    337117224 : '''IO error writing file: @1\n''', 
    337117225 : '''IO error seeking file: @1\n''', 
    337117226 : '''Error opening database file: @1\n''', 
    337117227 : '''Error in posix_fadvise(@1) for database @2\n''', 
    337117228 : '''Error creating database file: @1\n''', 
    337117229 : '''Error opening backup file: @1\n''', 
    337117230 : '''Error creating backup file: @1\n''', 
    337117231 : '''Unexpected end of database file @1\n''', 
    337117232 : '''Database @1 is not in state (@2) to be safely fixed up\n''', 
    337117233 : '''Database error\n''', 
    337117234 : '''Username or password is too long\n''', 
    337117235 : '''Cannot find record for database "@1" backup level @2 in the backup history\n''', 
    337117236 : '''Internal error. History query returned null SCN or GUID\n''', 
    337117237 : '''Unexpected end of file when reading header of database file "@1" (stage @2)\n''', 
    337117238 : '''Internal error. Database file is not locked. Flags are @1\n''', 
    337117239 : '''Internal error. Cannot get backup guid clumplet\n''', 
    337117240 : '''Internal error. Database page @1 had been changed during backup (page SCN=@2, backup SCN=@3)\n''', 
    337117241 : '''Database file size is not a multiple of page size\n''', 
    337117242 : '''Level 0 backup is not restored\n''', 
    337117243 : '''Unexpected end of file when reading header of backup file: @1\n''', 
    337117244 : '''Invalid incremental backup file: @1\n''', 
    337117245 : '''Unsupported version @1 of incremental backup file: @2\n''', 
    337117246 : '''Invalid level @1 of incremental backup file: @2, expected @3\n''', 
    337117247 : '''Wrong order of backup files or invalid incremental backup file detected, file: @1\n''', 
    337117248 : '''Unexpected end of backup file: @1\n''', 
    337117249 : '''Error creating database file: @1 via copying from: @2\n''', 
    337117250 : '''Unexpected end of file when reading header of restored database file (stage @1)\n''', 
    337117251 : '''Cannot get backup guid clumplet from L0 backup\n''', 
    337117255 : '''Wrong parameter @1 for switch -D, need ON or OFF\n''', 
    337117257 : '''Terminated due to user request\n''', 
    337182750 : '''conflicting actions "@1" and "@2" found\n''', 
    337182751 : '''action switch not found\n''', 
    337182752 : '''switch "@1" must be set only once\n''', 
    337182753 : '''value for switch "@1" is missing\n''', 
    337182754 : '''invalid value ("@1") for switch "@2"\n''', 
    337182755 : '''unknown switch "@1" encountered\n''', 
    337182756 : '''switch "@1" can be used by service only\n''', 
    337182757 : '''switch "@1" can be used by interactive user only\n''', 
    337182758 : '''mandatory parameter "@1" for switch "@2" is missing\n''', 
    337182759 : '''parameter "@1" is incompatible with action "@2"\n''', 
    337182760 : '''mandatory switch "@1" is missing\n''', 
}



##############################################################################
# Copyright (c) 2014 Hajime Nakagami<nakagami@gmail.com>
# All rights reserved.
# Licensed under the New BSD License
# (http://www.freebsd.org/copyright/freebsd-license.html)
#
# Python DB-API 2.0 module for Firebird. 
##############################################################################
import sys

PYTHON_MAJOR_VER = sys.version_info[0]

if PYTHON_MAJOR_VER == 3:
    def ord(c):
        return c

class Arc4:
    def __init__(self, key):
        state = list(range(256))
        index1 = 0
        index2 = 0

        for i in range(256):
            index2 = (ord(key[index1]) + state[i] + index2) % 256
            (state[i], state[index2]) = (state[index2], state[i])
            index1 = (index1 + 1) % len(key)

        self.state = state
        self.x = 0
        self.y = 0

    def translate(self, plain):
        state = self.state
        enc=b''
        for i in range(len(plain)):
            self.x = (self.x + 1) % 256
            self.y = (self.y + state[self.x]) % 256
            (state[self.x], state[self.y]) = (state[self.y], state[self.x])
            xorIndex = (state[self.x]+state[self.y]) % 256
            if PYTHON_MAJOR_VER == 3:
                enc += bytes([plain[i] ^ state[xorIndex]])
            else:
                enc += chr(ord(plain[i]) ^ state[xorIndex])
        return enc



##############################################################################
# Copyright (c) 2011-2014 Hajime Nakagami<nakagami@gmail.com>
# All rights reserved.
# Licensed under the New BSD License
# (http://www.freebsd.org/copyright/freebsd-license.html)
#
# Python DB-API 2.0 module for Firebird. 
##############################################################################
import sys

PYTHON_MAJOR_VER = sys.version_info[0]

DEFAULT_CHARSET='UTF8'

ISC_TIME_SECONDS_PRECISION = 10000
MAX_CHAR_LENGTH = 32767
BLOB_SEGMENT_SIZE = 32000

isc_info_end = 1
isc_info_truncated = 2
isc_info_error = 3
isc_info_data_not_ready = 4
isc_info_length = 126
isc_info_flag_end = 127

isc_info_db_id = 4
isc_info_reads = 5
isc_info_writes = 6
isc_info_fetches = 7
isc_info_marks = 8
isc_info_implementation = 11
isc_info_version = 12
isc_info_base_level = 13
isc_info_page_size = 14
isc_info_num_buffers = 15
isc_info_limbo = 16
isc_info_current_memory = 17
isc_info_max_memory = 18
isc_info_window_turns = 19
isc_info_license = 20
isc_info_allocation = 21
isc_info_attachment_id = 22
isc_info_read_seq_count = 23
isc_info_read_idx_count = 24
isc_info_insert_count = 25
isc_info_update_count = 26
isc_info_delete_count = 27
isc_info_backout_count = 28
isc_info_purge_count = 29
isc_info_expunge_count = 30
isc_info_sweep_interval = 31
isc_info_ods_version = 32
isc_info_ods_minor_version = 33
isc_info_no_reserve = 34
isc_info_logfile = 35
isc_info_cur_logfile_name = 36
isc_info_cur_log_part_offset = 37
isc_info_num_wal_buffers = 38
isc_info_wal_buffer_size = 39
isc_info_wal_ckpt_length = 40
isc_info_wal_cur_ckpt_interval = 41
isc_info_wal_prv_ckpt_fname = 42
isc_info_wal_prv_ckpt_poffset = 43
isc_info_wal_recv_ckpt_fname = 44
isc_info_wal_recv_ckpt_poffset = 45
isc_info_wal_grpc_wait_usecs = 47
isc_info_wal_num_io = 48
isc_info_wal_avg_io_size = 49
isc_info_wal_num_commits = 50
isc_info_wal_avg_grpc_size = 51
isc_info_forced_writes = 52
isc_info_user_names = 53
isc_info_page_errors = 54
isc_info_record_errors = 55
isc_info_bpage_errors = 56
isc_info_dpage_errors = 57
isc_info_ipage_errors = 58
isc_info_ppage_errors = 59
isc_info_tpage_errors = 60
isc_info_set_page_buffers = 61
isc_info_db_sql_dialect = 62
isc_info_db_read_only = 63
isc_info_db_size_in_pages = 64
isc_info_att_charset = 101
isc_info_db_class = 102
isc_info_firebird_version = 103
isc_info_oldest_transaction = 104
isc_info_oldest_active = 105
isc_info_oldest_snapshot = 106
isc_info_next_transaction = 107
isc_info_db_provider = 108
isc_info_active_transactions = 109
isc_info_active_tran_count = 110
isc_info_creation_date = 111
isc_info_db_file_size = 112

# isc_info_sql_records items
isc_info_req_select_count = 13
isc_info_req_insert_count = 14
isc_info_req_update_count = 15
isc_info_req_delete_count = 16

isc_info_svc_svr_db_info = 50
isc_info_svc_get_license = 51
isc_info_svc_get_license_mask = 52
isc_info_svc_get_config = 53
isc_info_svc_version = 54
isc_info_svc_server_version = 55
isc_info_svc_implementation = 56
isc_info_svc_capabilities = 57
isc_info_svc_user_dbpath = 58
isc_info_svc_get_env = 59
isc_info_svc_get_env_lock = 60
isc_info_svc_get_env_msg = 61
isc_info_svc_line = 62
isc_info_svc_to_eof = 63
isc_info_svc_timeout = 64
isc_info_svc_get_licensed_users = 65
isc_info_svc_limbo_trans = 66
isc_info_svc_running = 67
isc_info_svc_get_users = 68

SQL_TYPE_TEXT = 452
SQL_TYPE_VARYING = 448
SQL_TYPE_SHORT = 500
SQL_TYPE_LONG = 496
SQL_TYPE_FLOAT = 482
SQL_TYPE_DOUBLE = 480
SQL_TYPE_D_FLOAT = 530
SQL_TYPE_TIMESTAMP = 510
SQL_TYPE_BLOB = 520
SQL_TYPE_ARRAY = 540
SQL_TYPE_QUAD = 550
SQL_TYPE_TIME = 560
SQL_TYPE_DATE = 570
SQL_TYPE_INT64 = 580
SQL_TYPE_BOOLEAN = 32764
SQL_TYPE_NULL = 32766


ISOLATION_LEVEL_READ_COMMITED_LEGACY = 0
ISOLATION_LEVEL_READ_COMMITED = 1
ISOLATION_LEVEL_REPEATABLE_READ = 2
ISOLATION_LEVEL_SNAPSHOT = ISOLATION_LEVEL_REPEATABLE_READ
ISOLATION_LEVEL_SERIALIZABLE = 3
ISOLATION_LEVEL_READ_COMMITED_RO = 4

# Database Parameter Block parameter
isc_dpb_version1 = 1
isc_dpb_version2 = 2
isc_dpb_cdd_pathname = 1
isc_dpb_allocation = 2
isc_dpb_journal = 3
isc_dpb_page_size = 4
isc_dpb_num_buffers = 5
isc_dpb_buffer_length = 6
isc_dpb_debug = 7
isc_dpb_garbage_collect = 8
isc_dpb_verify = 9
isc_dpb_sweep = 10
isc_dpb_enable_journal = 11
isc_dpb_disable_journal = 12
isc_dpb_dbkey_scope = 13
isc_dpb_number_of_users = 14
isc_dpb_trace = 15
isc_dpb_no_garbage_collect = 16
isc_dpb_damaged = 17
isc_dpb_license = 18
isc_dpb_sys_user_name = 19
isc_dpb_encrypt_key = 20
isc_dpb_activate_shadow = 21
isc_dpb_sweep_interval = 22
isc_dpb_delete_shadow = 23
isc_dpb_force_write = 24
isc_dpb_begin_log = 25
isc_dpb_quit_log = 26
isc_dpb_no_reserve = 27
isc_dpb_user_name = 28
isc_dpb_password = 29
isc_dpb_password_enc = 30
isc_dpb_sys_user_name_enc = 31
isc_dpb_interp = 32
isc_dpb_online_dump = 33
isc_dpb_old_file_size = 34
isc_dpb_old_num_files = 35
isc_dpb_old_file = 36
isc_dpb_old_start_page = 37
isc_dpb_old_start_seqno = 38
isc_dpb_old_start_file = 39
isc_dpb_drop_walfile = 40
isc_dpb_old_dump_id = 41
isc_dpb_wal_backup_dir = 42
isc_dpb_wal_chkptlen = 43
isc_dpb_wal_numbufs = 44
isc_dpb_wal_bufsize = 45
isc_dpb_wal_grp_cmt_wait = 46
isc_dpb_lc_messages = 47
isc_dpb_lc_ctype = 48
isc_dpb_cache_manager = 49
isc_dpb_shutdown = 50
isc_dpb_online = 51
isc_dpb_shutdown_delay = 52
isc_dpb_reserved = 53
isc_dpb_overwrite = 54
isc_dpb_sec_attach = 55
isc_dpb_disable_wal = 56
isc_dpb_connect_timeout = 57
isc_dpb_dummy_packet_interval = 58
isc_dpb_gbak_attach = 59
isc_dpb_sql_role_name = 60
isc_dpb_set_page_buffers = 61
isc_dpb_working_directory = 62
isc_dpb_sql_dialect = 63
isc_dpb_set_db_readonly = 64
isc_dpb_set_db_sql_dialect = 65
isc_dpb_gfix_attach = 66
isc_dpb_gstat_attach = 67
isc_dpb_set_db_charset = 68
isc_dpb_gsec_attach = 69
isc_dpb_address_path = 70
isc_dpb_process_id = 71
isc_dpb_no_db_triggers = 72
isc_dpb_trusted_auth = 73
isc_dpb_process_name = 74
isc_dpb_trusted_role = 75
isc_dpb_org_filename = 76
isc_dpb_utf8_filename = 77
isc_dpb_ext_call_depth = 78
isc_dpb_auth_block = 79
isc_dpb_client_version = 80
isc_dpb_remote_protocol = 81
isc_dpb_host_name = 82
isc_dpb_os_user	 = 83
isc_dpb_specific_auth_data = 84
isc_dpb_auth_plugin_list = 85
isc_dpb_auth_plugin_name = 86
isc_dpb_config = 87
isc_dpb_nolinger = 88

# Transaction Parameter Block parameter
isc_tpb_version1 = 1
isc_tpb_version3 = 3
isc_tpb_consistency = 1
isc_tpb_concurrency = 2
isc_tpb_shared = 3
isc_tpb_protected = 4
isc_tpb_exclusive = 5
isc_tpb_wait = 6
isc_tpb_nowait = 7
isc_tpb_read = 8
isc_tpb_write = 9
isc_tpb_lock_read = 10
isc_tpb_lock_write = 11
isc_tpb_verb_time = 12
isc_tpb_commit_time = 13
isc_tpb_ignore_limbo = 14
isc_tpb_read_committed = 15
isc_tpb_autocommit = 16
isc_tpb_rec_version = 17
isc_tpb_no_rec_version = 18
isc_tpb_restart_requests = 19
isc_tpb_no_auto_undo = 20
isc_tpb_lock_timeout = 21


# Service Parameter Block parameter
isc_spb_version1 = 1
isc_spb_current_version = 2
isc_spb_version = isc_spb_current_version
isc_spb_user_name = 28              # isc_dpb_user_name
isc_spb_sys_user_name = 19          # isc_dpb_sys_user_name
isc_spb_sys_user_name_enc = 31      # isc_dpb_sys_user_name_enc
isc_spb_password = 29               # isc_dpb_password
isc_spb_password_enc = 30           # isc_dpb_password_enc
isc_spb_command_line = 105
isc_spb_dbname = 106
isc_spb_verbose = 107
isc_spb_options = 108
isc_spb_address_path = 109
isc_spb_process_id = 110
isc_spb_trusted_auth = 111
isc_spb_process_name = 112
isc_spb_trusted_role = 113
isc_spb_connect_timeout = 57        # isc_dpb_connect_timeout
isc_spb_dummy_packet_interval = 58  # isc_dpb_dummy_packet_interval
isc_spb_sql_role_name = 60          # isc_dpb_sql_role_name

# isc_action_svc_properties params
isc_spb_prp_page_buffers = 5
isc_spb_prp_sweep_interval = 6
isc_spb_prp_shutdown_db = 7
isc_spb_prp_deny_new_attachments = 9
isc_spb_prp_deny_new_transactions = 10
isc_spb_prp_reserve_space = 11
isc_spb_prp_write_mode = 12
isc_spb_prp_access_mode = 13
isc_spb_prp_set_sql_dialect = 14
isc_spb_prp_activate = 0x0100
isc_spb_prp_db_online = 0x0200
isc_spb_prp_force_shutdown = 41
isc_spb_prp_attachments_shutdown = 42
isc_spb_prp_transactions_shutdown = 43
isc_spb_prp_shutdown_mode = 44
isc_spb_prp_online_mode = 45

# backup
isc_spb_bkp_file = 5
isc_spb_bkp_factor = 6
isc_spb_bkp_length = 7
isc_spb_bkp_ignore_checksums = 0x01
isc_spb_bkp_ignore_limbo = 0x02
isc_spb_bkp_metadata_only = 0x04
isc_spb_bkp_no_garbage_collect = 0x08
isc_spb_bkp_old_descriptions = 0x10
isc_spb_bkp_non_transportable = 0x20
isc_spb_bkp_convert = 0x40
isc_spb_bkp_expand = 0x8

# restore
isc_spb_res_buffers = 9
isc_spb_res_page_size = 10
isc_spb_res_length = 11
isc_spb_res_access_mode = 12
isc_spb_res_deactivate_idx = 0x0100
isc_spb_res_no_shadow = 0x0200
isc_spb_res_no_validity = 0x0400
isc_spb_res_one_at_a_time = 0x0800
isc_spb_res_replace = 0x1000
isc_spb_res_create = 0x2000
isc_spb_res_use_all_space = 0x4000

# trace
isc_spb_trc_id = 1
isc_spb_trc_name = 2
isc_spb_trc_cfg = 3

# isc_info_svc_svr_db_info params
isc_spb_num_att = 5
isc_spb_num_db = 6

# isc_info_svc_db_stats params
isc_spb_sts_data_pages = 0x01
isc_spb_sts_db_log =0x02
isc_spb_sts_hdr_pages = 0x04
isc_spb_sts_idx_pages = 0x08
isc_spb_sts_sys_relations = 0x10
isc_spb_sts_record_versions = 0x20
isc_spb_sts_table = 0x40
isc_spb_sts_nocreation = 0x80

# isc_action_svc_repair params
isc_spb_rpr_validate_db = 0x01
isc_spb_rpr_sweep_db = 0x02
isc_spb_rpr_mend_db = 0x04
isc_spb_rpr_list_limbo_trans = 0x08
isc_spb_rpr_check_db = 0x10
isc_spb_rpr_ignore_checksum = 0x20
isc_spb_rpr_kill_shadows = 0x40
isc_spb_rpr_full = 0x80

# Service Action Items
isc_action_svc_backup = 1
isc_action_svc_restore = 2
isc_action_svc_repair = 3
isc_action_svc_add_user = 4
isc_action_svc_delete_user = 5
isc_action_svc_modify_user = 6
isc_action_svc_display_user = 7
isc_action_svc_properties = 8
isc_action_svc_add_license = 9
isc_action_svc_remove_license = 10
isc_action_svc_db_stats = 11
isc_action_svc_get_ib_log = 12
isc_action_svc_get_fb_log = 12
isc_action_svc_nbak = 20
isc_action_svc_nrest = 21
isc_action_svc_trace_start = 22
isc_action_svc_trace_stop = 23
isc_action_svc_trace_suspend = 24
isc_action_svc_trace_resume = 25
isc_action_svc_trace_list = 26
isc_action_svc_set_mapping = 27
isc_action_svc_drop_mapping = 28
isc_action_svc_display_user_adm = 29
isc_action_svc_last = 30

# Transaction informatino items
isc_info_tra_id = 4
isc_info_tra_oldest_interesting = 5
isc_info_tra_oldest_snapshot = 6
isc_info_tra_oldest_active = 7
isc_info_tra_isolation = 8
isc_info_tra_access = 9
isc_info_tra_lock_timeout = 10

# SQL information items
isc_info_sql_select = 4
isc_info_sql_bind = 5
isc_info_sql_num_variables = 6
isc_info_sql_describe_vars = 7
isc_info_sql_describe_end = 8
isc_info_sql_sqlda_seq = 9
isc_info_sql_message_seq = 10
isc_info_sql_type = 11
isc_info_sql_sub_type = 12
isc_info_sql_scale = 13
isc_info_sql_length = 14
isc_info_sql_null_ind = 15
isc_info_sql_field = 16
isc_info_sql_relation = 17
isc_info_sql_owner = 18
isc_info_sql_alias = 19
isc_info_sql_sqlda_start = 20
isc_info_sql_stmt_type = 21
isc_info_sql_get_plan = 22
isc_info_sql_records = 23
isc_info_sql_batch_fetch = 24

isc_info_sql_stmt_select = 1
isc_info_sql_stmt_insert = 2
isc_info_sql_stmt_update = 3
isc_info_sql_stmt_delete = 4
isc_info_sql_stmt_ddl = 5
isc_info_sql_stmt_get_segment = 6
isc_info_sql_stmt_put_segment = 7
isc_info_sql_stmt_exec_procedure = 8
isc_info_sql_stmt_start_trans = 9
isc_info_sql_stmt_commit = 10
isc_info_sql_stmt_rollback = 11
isc_info_sql_stmt_select_for_upd = 12
isc_info_sql_stmt_set_generator = 13
isc_info_sql_stmt_savepoint = 14

isc_arg_end = 0
isc_arg_gds = 1
isc_arg_string = 2
isc_arg_cstring = 3
isc_arg_number = 4
isc_arg_interpreted = 5
isc_arg_vms = 6
isc_arg_unix = 7
isc_arg_domain = 8
isc_arg_dos = 9
isc_arg_mpexl = 10
isc_arg_mpexl_ipc = 11
isc_arg_next_mach = 15
isc_arg_netware = 16
isc_arg_win32 = 17
isc_arg_warning = 18
isc_arg_sql_state = 19

# Protocol Types (accept_type)
ptype_batch_send = 3    # Batch sends, no asynchrony
ptype_out_of_band = 4   # Batch sends w/ out of band notification
ptype_lazy_send = 5     # Deferred packets delivery

PROTOCOL_VERSION10 = 10
PROTOCOL_VERSION11 = 11
PROTOCOL_VERSION12 = 12
PROTOCOL_VERSION13 = 13

CNCT_user = 1
CNCT_passwd = 2
CNCT_host = 4
CNCT_group = 5
CNCT_user_verification = 6
CNCT_specific_data = 7
CNCT_plugin_name = 8
CNCT_login = 9
CNCT_plugin_list = 10
CNCT_client_crypt = 11

DSQL_close = 1
DSQL_drop = 2

charset_map = {
    # DB CHAR SET NAME    :   PYTHON CODEC NAME (CANONICAL)
    # --------------------------------------------------------------------------
    'OCTETS'              :   None, # Allow to pass through unchanged.
    'UNICODE_FSS'         :   'utf_8',
    'UTF8'                :   'utf_8', # (Firebird 2.0+)
    'SJIS_0208'           :   'shift_jis',
    'EUCJ_0208'           :   'euc_jp',
    'DOS737'              :   'cp737',
    'DOS437'              :   'cp437',
    'DOS850'              :   'cp850',
    'DOS865'              :   'cp865',
    'DOS860'              :   'cp860',
    'DOS863'              :   'cp863',
    'DOS775'              :   'cp775',
    'DOS862'              :   'cp862',
    'DOS864'              :   'cp864',
    'ISO8859_1'           :   'iso8859_1',
    'ISO8859_2'           :   'iso8859_2',
    'ISO8859_3'           :   'iso8859_3',
    'ISO8859_4'           :   'iso8859_4',
    'ISO8859_5'           :   'iso8859_5',
    'ISO8859_6'           :   'iso8859_6',
    'ISO8859_7'           :   'iso8859_7',
    'ISO8859_8'           :   'iso8859_8',
    'ISO8859_9'           :   'iso8859_9',
    'ISO8859_13'          :   'iso8859_13',
    'KSC_5601'            :   'euc_kr',
    'DOS852'              :   'cp852',
    'DOS857'              :   'cp857',
    'DOS861'              :   'cp861',
    'DOS866'              :   'cp866',
    'DOS869'              :   'cp869',
    'WIN1250'             :   'cp1250',
    'WIN1251'             :   'cp1251',
    'WIN1252'             :   'cp1252',
    'WIN1253'             :   'cp1253',
    'WIN1254'             :   'cp1254',
    'BIG_5'               :   'big5',
    'GB_2312'             :   'gb2312',
    'WIN1255'             :   'cp1255',
    'WIN1256'             :   'cp1256',
    'WIN1257'             :   'cp1257',
    'KOI8-R'              :   'koi8_r', # (Firebird 2.0+)
    'KOI8-U'              :   'koi8_u', # (Firebird 2.0+)
    'WIN1258'             :   'cp1258', # (Firebird 2.0+)
}



##############################################################################
# Copyright (c) 2009-2014 Hajime Nakagami<nakagami@gmail.com>
# All rights reserved.
# Licensed under the New BSD License
# (http://www.freebsd.org/copyright/freebsd-license.html)
#
# Python DB-API 2.0 module for Firebird.
##############################################################################
from __future__ import print_function
import sys
import warnings
import xdrlib, time, datetime, decimal, struct
import itertools
from collections import Mapping
from firebirdsql.fberrmsgs import messages
from firebirdsql import (DatabaseError, InternalError, OperationalError,
    ProgrammingError, IntegrityError, DataError, NotSupportedError,)
from firebirdsql.consts import *
from firebirdsql.utils import *
from firebirdsql.wireprotocol import WireProtocol, INFO_SQL_SELECT_DESCRIBE_VARS
from firebirdsql.socketstream import SocketStream
from firebirdsql.xsqlvar import XSQLVAR, calc_blr, parse_select_items, parse_xsqlda
__version__ = '0.9.2'
apilevel = '2.0'
threadsafety = 1
paramstyle = 'qmark'

DEBUG = False

def DEBUG_OUTPUT(*argv):
    if not DEBUG:
        return
    for s in argv:
        print(s, end=' ', file=sys.stderr)
    print(file=sys.stderr)

transaction_parameter_block = (
    # ISOLATION_LEVEL_READ_COMMITED_LEGACY
    bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_no_rec_version]),
    # ISOLATION_LEVEL_READ_COMMITED
    bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_read_committed, isc_tpb_rec_version]),
    # ISOLATION_LEVEL_REPEATABLE_READ
    bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_concurrency]),
    # ISOLATION_LEVEL_SERIALIZABLE
    bs([isc_tpb_version3, isc_tpb_write, isc_tpb_wait, isc_tpb_consistency]),
    # ISOLATION_LEVEL_READ_COMMITED_RO
    bs([isc_tpb_version3, isc_tpb_read, isc_tpb_wait, isc_tpb_read_committed,isc_tpb_rec_version]),
)

Date = datetime.date
Time = datetime.time
TimeDelta = datetime.timedelta
Timestamp = datetime.datetime
def DateFromTicks(ticks):
    return apply(Date,time.localtime(ticks)[:3])
def TimeFromTicks(ticks):
    return apply(Time,time.localtime(ticks)[3:6])
def TimestampFromTicks(ticks):
    return apply(Timestamp,time.localtime(ticks)[:6])
def Binary(b):
    return bytes(b)

class DBAPITypeObject:
    def __init__(self,*values):
        self.values = values
    def __cmp__(self,other):
        if other in self.values:
            return 0
        if other < self.values:
            return 1
        else:
            return -1
STRING = DBAPITypeObject(str)
if PYTHON_MAJOR_VER==3:
    BINARY = DBAPITypeObject(bytes)
else:
    BINARY = DBAPITypeObject(str)
NUMBER = DBAPITypeObject(int, decimal.Decimal)
DATETIME = DBAPITypeObject(datetime.datetime, datetime.date, datetime.time)
DATE = DBAPITypeObject(datetime.date)
TIME = DBAPITypeObject(datetime.time)
ROWID = DBAPITypeObject()

class Statement(object):
    """
    statement handle and status (open/close)
    """
    def __init__(self, trans):
        DEBUG_OUTPUT("Statement::__init__()")
        self.trans = trans
        self._allocate_stmt()
        self._is_open = False
        self.stmt_type = None

    def _allocate_stmt(self):
        self.trans.connection._op_allocate_statement()
        if self.trans.connection.accept_type == ptype_lazy_send:
            self.handle = -1
        else:
            (h, oid, buf) = self.trans.connection._op_response()
            self.handle = h

    def prepare(self, sql, explain_plan=False):
        DEBUG_OUTPUT("Statement::prepare()", self.handle)
        if explain_plan:
            self.trans.connection._op_prepare_statement(
                self.handle, self.trans.trans_handle, sql,
                option_items=bs([isc_info_sql_get_plan]))
        else:
            self.trans.connection._op_prepare_statement(
                self.handle, self.trans.trans_handle, sql)
            self.plan = None

        (h, oid, buf) = self.trans.connection._op_response()

        i = 0
        if byte_to_int(buf[i]) == isc_info_sql_get_plan:
            l = bytes_to_int(buf[i+1:i+3])
            self.plan = self.trans.connection.bytes_to_str(buf[i+3:i+3+l])
            i += 3 + l
        self.stmt_type, self.xsqlda = parse_xsqlda(buf[i:],
                                    self.trans.connection, self.handle)

    def open(self):
        DEBUG_OUTPUT("Statement::open()")
        self._is_open = True

    def close(self):
        DEBUG_OUTPUT("Statement::close()", self.handle)
        if (self.stmt_type == isc_info_sql_stmt_select and self._is_open):
            self.trans.connection._op_free_statement(self.handle, DSQL_close)
            if self.trans.connection.accept_type != ptype_lazy_send:
                (h, oid, buf) = self.trans.connection._op_response()
        self._is_open = False

    def drop(self):
        DEBUG_OUTPUT("Statement::drop()")
        if self.handle != -1:
            self.trans.connection._op_free_statement(self.handle, DSQL_drop)
            if self.trans.connection.accept_type != ptype_lazy_send:
                (h, oid, buf) = self.trans.connection._op_response()
        self._is_open = False
        self.handle = -1

    @property
    def is_opened(self):
        return self._is_open and self.handle != -1

class PreparedStatement(object):
    def __init__(self, cur, sql, explain_plan=False):
        DEBUG_OUTPUT("PreparedStatement::__init__()")
        self.cur = cur
        self.sql = sql
        transaction = self.cur.transaction
        connection = transaction.connection

        self.stmt = Statement(transaction)
        self.stmt.prepare(sql, explain_plan)

    def __getattr__(self, attrname):
        if attrname == 'description':
            if len(self.stmt.xsqlda) == 0:
                return None
            r = []
            for x in self.stmt.xsqlda:
                r.append((x.aliasname, x.sqltype, x.display_length(),
                        x.io_length(), x.precision(),
                        x.sqlscale, True if x.null_ok else False))
            return r
        elif attrname == 'n_output_params':
            return len(self.stmt.xsqlda)
        raise AttributeError


def _fetch_generator(stmt):
    DEBUG_OUTPUT("_fetch_generator()", stmt.trans._trans_handle)
    connection = stmt.trans.connection
    more_data = True
    while more_data:
        if not stmt.is_opened:
            raise StopIteration()
        connection._op_fetch(stmt.handle, calc_blr(stmt.xsqlda))
        (rows, more_data) = connection._op_fetch_response(
                                        stmt.handle, stmt.xsqlda)
        for r in rows:
            # Convert BLOB handle to data
            for i in range(len(stmt.xsqlda)):
                x = stmt.xsqlda[i]
                if x.sqltype == SQL_TYPE_BLOB:
                    if not r[i]:
                        continue
                    connection._op_open_blob(r[i], stmt.trans.trans_handle)
                    (h, oid, buf) = connection._op_response()
                    v = bs([])
                    n = 1   # 1:mora data 2:no more data
                    while n == 1:
                        connection._op_get_segment(h)
                        (n, oid, buf) = connection._op_response()
                        while buf:
                            ln = bytes_to_int(buf[:2])
                            v += buf[2:ln+2]
                            buf = buf[ln+2:]
                    connection._op_close_blob(h)
                    if connection.accept_type != ptype_lazy_send:
                        (h, oid, buf) = connection._op_response()
                    r[i] = v
                    if x.sqlsubtype == 1:    # TEXT
                        if connection.use_unicode:
                            r[i] = connection.bytes_to_ustr(r[i])
                        else:
                            r[i] = connection.bytes_to_str(r[i])
            yield tuple(r)
    DEBUG_OUTPUT("_fetch_generator() StopIteration()")
    raise StopIteration()

class Cursor(object):
    def __init__(self, trans):
        DEBUG_OUTPUT("Cursor::__init__()")
        self._transaction = trans
        self.stmt = None
        self.arraysize = 1

    def __enter__(self):
        return self

    def __exit__(self, exc, value, traceback):
        self.close()

    @property
    def transaction(self):
        return self._transaction

    def _convert_params(self, params):
        cooked_params = []
        for param in params:
            if (type(param)==str or
                (PYTHON_MAJOR_VER == 2 and type(param)==unicode)):
                param = self.transaction.connection.str_to_bytes(param)
            cooked_params.append(param)
        return cooked_params

    def _get_stmt(self, query):
        self.query = query
        if isinstance(query, PreparedStatement):
            stmt = query.stmt
            stmt.close()
        else:
            if self.stmt:
                self.stmt.close()
            else:
                self.stmt = Statement(self.transaction)
            stmt = self.stmt
            stmt.prepare(query)
        return stmt

    def prep(self, query, explain_plan=False):
        DEBUG_OUTPUT("Cursor::prep()")
        prepared_statement = PreparedStatement(self, query,
                                                explain_plan=explain_plan)
        return prepared_statement

    def execute(self, query, params=[]):
        DEBUG_OUTPUT("Cursor::execute()", query, params)
        stmt = self._get_stmt(query)
        cooked_params = self._convert_params(params)
        if stmt.stmt_type == isc_info_sql_stmt_exec_procedure:
            self.transaction.connection._op_execute2(stmt.handle,
                self.transaction.trans_handle, cooked_params,
                calc_blr(stmt.xsqlda))
            self._callproc_result = \
                self.transaction.connection._op_sql_response(stmt.xsqlda)
            self.transaction.connection._op_response()
            self._fetch_records = None
        else:
            self.transaction.connection._op_execute(stmt.handle,
                                self.transaction.trans_handle, cooked_params)
            try:
                (h, oid, buf) = self.transaction.connection._op_response()
            except OperationalError as e:
                self._fetch_records = None
                self._callproc_result = None
                if e.gds_codes.intersection(
                    [335544665, 335544466, 335544838, 335544347]):
                    raise IntegrityError(e._message, e.gds_codes, e.sql_code)
                if e.sql_code == -303:
                    warnings.warn(e._message)
                    return
                else:
                    raise OperationalError(e._message, e.gds_codes, e.sql_code)

            if stmt.stmt_type == isc_info_sql_stmt_select:
                self._fetch_records = _fetch_generator(stmt)
                stmt.open()
            else:
                self._fetch_records = None
            self._callproc_result = None
        self.transaction.is_dirty = True

    def callproc(self, procname, params=[]):
        DEBUG_OUTPUT("Cursor::callproc()")
        query = 'EXECUTE PROCEDURE ' + procname + ' ' + ','.join('?'*len(params))
        return self.execute(query, params)

    def executemany(self, query, seq_of_params):
        for params in seq_of_params:
            self.execute(query, params)


    def fetchone(self):
        if not self.transaction.is_dirty:
            return None
        # callproc or not select statement
        if not self._fetch_records:
            if self._callproc_result:
                r = self._callproc_result
                self._callproc_result = None
                return r
            return None
        # select statement
        try:
            if PYTHON_MAJOR_VER == 3:
                return tuple(next(self._fetch_records))
            else:
                return tuple(self._fetch_records.next())
        except StopIteration:
            return None

    def __iter__(self):
        return self

    def __next__(self):
        r = self.fetchone()
        if not r:
            raise StopIteration()
        return r

    def next(self):
        return self.__next__()

    def fetchall(self):
        # callproc or not select statement
        if not self.transaction.is_dirty:
            return None
        if not self._fetch_records:
            if self._callproc_result:
                r = [tuple(self._callproc_result)]
                self._callproc_result = None
                return r
            return []
        # select statement
        return [tuple(r) for r in self._fetch_records]

    def fetchmany(self, size=None):
        if not size:
            size = self.arraysize
        # callproc or not select statement
        if not self._fetch_records:
            if self._callproc_result:
                r = [self._callproc_result]
                self._callproc_result = None
                return r
            return []
        # select statement
        return list(itertools.islice(self._fetch_records, size))

    # kinterbasdb extended API
    def fetchonemap(self):
        r = self.fetchone()
        if r is None:
            return {}
        return RowMapping(r, self.description)

    def fetchallmap(self):
        desc = self.description
        return [RowMapping(row, desc) for row in self.fetchall()]

    def fetchmanymap(self, size=None):
        desc = self.description
        return [RowMapping(row, desc) for row in self.fetchmany(size)]

    def itermap(self):
        r = self.fetchonemap()
        while r:
            yield r
            r = self.fetchonemap()

    def close(self):
        DEBUG_OUTPUT("Cursor::close()")
        if not self.stmt:
            return
        self.stmt.drop()
        self.stmt = None

    def nextset(self):
        raise NotSupportedError()

    def setinputsizes(self, sizes):
        pass

    def setoutputsize(self, size, column):
        pass

    @property
    def description(self):
        if not self.stmt:
            return None
        return [(x.aliasname, x.sqltype, x.display_length(), x.io_length(),
                    x.precision(), x.sqlscale, True if x.null_ok else False)
                for x in self.stmt.xsqlda]

    @property
    def rowcount(self):
        DEBUG_OUTPUT("Cursor::rowcount()")
        if self.stmt.handle == -1:
            return -1

        self.transaction.connection._op_info_sql(self.stmt.handle,
                                     bs([isc_info_sql_records]))
        (h, oid, buf) = self.transaction.connection._op_response()
        assert buf[:3] == bs([0x17,0x1d,0x00]) # isc_info_sql_records
        if self.stmt.stmt_type == isc_info_sql_stmt_select:
            assert buf[17:20] == bs([0x0d,0x04,0x00])
                                              # isc_info_req_select_count
            count = bytes_to_int(buf[20:24])
        elif self.stmt.stmt_type == isc_info_sql_stmt_insert:
            assert buf[24:27] == bs([0x0e,0x04,0x00])
                                              # isc_info_req_insert_count
            count = bytes_to_int(buf[27:31])
        elif self.stmt.stmt_type == isc_info_sql_stmt_update:
            assert buf[3:6] == bs([0x0f,0x04,0x00])
                                              # isc_info_req_update_count
            count = bytes_to_int(buf[6:10])
        elif self.stmt.stmt_type == isc_info_sql_stmt_delete:
            assert buf[10:13] == bs([0x10,0x04,0x00])
                                              # isc_info_req_delete_count
            count = bytes_to_int(buf[13:17])
        DEBUG_OUTPUT("Cursor::rowcount()", self.stmt.stmt_type, count)
        return count

class EventConduit(WireProtocol):
    def __init__(self, conn, names, timeout):
        self.connection = conn
        self.event_names = {}
        for name in names:
            self.event_names[name] = 0
        self.timeout = timeout
        (h, port, family, ip_address) = self.connection._op_connect_request()
        self.sock = SocketStream(ip_address, port, timeout)
        self.connection.last_event_id += 1
        self.event_id = self.connection.last_event_id

        self.connection._op_que_events(self.event_names, 0, 0, self.event_id)
        (h, oid, buf) = self.connection._op_response()

        (event_id, event_names) = self._wait_for_event(timeout=timeout)
        assert event_id == self.event_id   # treat only one event_id
        self.event_names.update(event_names)

    def wait(self, timeout=None):
        self.connection._op_que_events(self.event_names, 0, 0, self.event_id)
        (h, oid, buf) = self.connection._op_response()

        r = self._wait_for_event(timeout=timeout)
        if r:
            (event_id, event_names) = r
            assert event_id == self.event_id   # treat only one event_id
            r = {}
            for k in event_names:
                r[k] = event_names[k]-self.event_names[k]
                self.event_names[k] = event_names[k]
        else:
            r = {}
            for k in self.event_names:
                r[k] = 0
        return r

    def close(self):
        self.connection._op_cancel_events(self.event_id)
        (h, oid, buf) = self.connection._op_response()
        self.sock.close()
        self.sock = None

class Connection(WireProtocol):
    def cursor(self):
        DEBUG_OUTPUT("Connection::cursor()")
        if self._transaction is None:
            self.begin()
        return Cursor(self._transaction)

    def begin(self):
        DEBUG_OUTPUT("Connection::begin()")
        if not self.sock:
            raise InternalError
        if self._transaction is None:
            self._transaction = Transaction(self, self._autocommit)
        self._transaction.begin()

    def commit(self, retaining=False):
        DEBUG_OUTPUT("Connection::commit()")
        if self._transaction:
            self._transaction.commit(retaining=retaining)

    def savepoint(self, name):
        return self._transaction.savepoint(name)

    def rollback(self, retaining=False, savepoint=None):
        DEBUG_OUTPUT("Connection::rollback()")
        if self._transaction:
            self._transaction.rollback(retaining=retaining, savepoint=savepoint)

    def execute_immediate(self, query):
        if self._transaction is None:
            self._transaction = Transaction(self, self._autocommit)
            self._transaction.begin()
        self._op_exec_immediate(
            self._transaction.trans_handle, query=query)
        (h, oid, buf) = self._op_response()
        self._transaction.is_dirty=True

    def __init__(self, dsn=None, user=None, password=None, role=None, host=None,
                    database=None, charset=DEFAULT_CHARSET, port=3050,
                    page_size=4096, is_services=False, cloexec=False,
                    timeout=None, isolation_level=None, use_unicode=None,
                    auth_plugin_list=('Srp', 'Legacy_Auth'),
                    wire_crypt=True, create_new=False):
        DEBUG_OUTPUT("Connection::__init__()")
        self.db_handle = None
        if dsn:
            i = dsn.find(':')
            if i < 0:
                self.hostname = host
                self.filename = dsn
            else:
                hostport = dsn[:i]
                self.filename = dsn[i+1:]
                i = hostport.find('/')
                if i < 0:
                    self.hostname = hostport
                else:
                    self.hostname = hostport[:i]
                    port = int(hostport[i+1:])
        else:
            self.hostname = host
            self.filename = database
        if self.hostname is None:
            self.hostname = 'localhost'
        self.port = port
        self.user = user
        self.password = password
        self.role = role
        self.charset = charset
        self.timeout = float(timeout) if timeout is not None else None
        self.auth_plugin_list = auth_plugin_list
        self.wire_crypt = wire_crypt
        self.page_size = page_size
        self.is_services = is_services
        if isolation_level is None:
            self.isolation_level = ISOLATION_LEVEL_READ_COMMITED
        else:
            self.isolation_level = int(isolation_level)
        self.use_unicode = use_unicode
        self.last_event_id = 0

        self._autocommit = False
        self._transaction = None
        self.sock = SocketStream(self.hostname, self.port, self.timeout, cloexec)

        self._op_connect(auth_plugin_list, wire_crypt)
        try:
            self._op_accept()
        except OperationalError as e:
            self.sock.close()
            self.sock = None
            raise e
        if create_new:                      # create database
            self._op_create(self.page_size)
        elif self.is_services:                  # service api
            self._op_service_attach()
        else:                                   # connect
            self._op_attach()
        (h, oid, buf) = self._op_response()
        self.db_handle = h

    def __enter__(self):
        return self

    def __exit__(self, exc, value, traceback):
        self.close()

    def set_isolation_level(self, isolation_level):
        self.isolation_level = int(isolation_level)

    def set_autocommit(self, is_autocommit):
        if self._autocommit != is_autocommit and self._transaction is not None:
            self.rollback()
            self._transaction = None
        self._autocommit = is_autocommit

    def _db_info(self, info_requests):
        if info_requests[-1] == isc_info_end:
            self._op_info_database(bs(info_requests))
        else:
            self._op_info_database(
                        bs(info_requests+type(info_requests)([isc_info_end])))
        (h, oid, buf) = self._op_response()
        i = 0
        i_request = 0
        r = []
        while i < len(buf):
            req = byte_to_int(buf[i])
            if req == isc_info_end:
                break
            assert req == info_requests[i_request] or req == isc_info_error
            if req == isc_info_user_names:
                user_names = []
                while req == isc_info_user_names:
                    l = bytes_to_int(buf[i+1:i+3])
                    user_names.append(buf[i+3:i+3+l])
                    i = i + 3 + l
                    req = byte_to_int(buf[i])
                r.append((req, user_names))
            else:
                l = bytes_to_int(buf[i+1:i+3])
                r.append((req, buf[i+3:i+3+l]))
                i = i + 3 + l
            i_request += 1
        return r

    def _db_info_convert_type(self, info_request, v):
        REQ_INT = set([
            isc_info_allocation, isc_info_no_reserve, isc_info_db_sql_dialect,
            isc_info_ods_minor_version, isc_info_ods_version,
            isc_info_page_size, isc_info_current_memory, isc_info_forced_writes,
            isc_info_max_memory, isc_info_num_buffers, isc_info_sweep_interval,
            isc_info_limbo, isc_info_attachment_id, isc_info_fetches,
            isc_info_marks, isc_info_reads, isc_info_writes,
            isc_info_set_page_buffers, isc_info_db_read_only,
            isc_info_db_size_in_pages, isc_info_page_errors,
            isc_info_record_errors, isc_info_bpage_errors,
            isc_info_dpage_errors, isc_info_ipage_errors,
            isc_info_ppage_errors, isc_info_tpage_errors,
            # may not be available in some versions of Firebird
            isc_info_oldest_transaction, isc_info_oldest_active,
            isc_info_oldest_snapshot, isc_info_next_transaction,
            isc_info_active_tran_count
        ])
        REQ_COUNT = set([
            isc_info_backout_count, isc_info_delete_count,
            isc_info_expunge_count, isc_info_insert_count, isc_info_purge_count,
            isc_info_read_idx_count, isc_info_read_seq_count,
            isc_info_update_count
        ])

        if info_request in (isc_info_base_level, ):
            # IB6 API guide p52
            return byte_to_int(v[1])
        elif info_request in (isc_info_db_id, ):
            # IB6 API guide p52
            conn_code = byte_to_int(v[0])
            len1 = byte_to_int(v[1])
            filename = self.bytes_to_str(v[2:2+len1])
            len2 = byte_to_int(v[2+len1])
            sitename = self.bytes_to_str(v[3+len1:3+len1+len2])
            return (conn_code, filename, sitename)
        elif info_request in (isc_info_implementation, ):
            return (byte_to_int(v[1]), byte_to_int(v[2]))
        elif info_request in (isc_info_version, isc_info_firebird_version):
            # IB6 API guide p53
            return self.bytes_to_str(v[2:2+byte_to_int(v[1])])
        elif info_request in (isc_info_user_names, ):
            # IB6 API guide p54
            user_names = []
            for u in v:
                user_names.append(self.bytes_to_str(u[1:]))
            return user_names
        elif info_request in REQ_INT:
            return bytes_to_int(v)
        elif info_request in REQ_COUNT:
            counts = {}
            i = 0
            while i < len(v):
                counts[bytes_to_int(v[i:i+2])] = bytes_to_int(v[i+2:i+6])
                i += 6
            return counts
        elif info_request in (isc_info_creation_date,):
            nday = bytes_to_int(v[:4]) + 2400001 - 1721119
            century = (4 * nday - 1) // 146097
            nday = 4 * nday - 1 - 146097 * century
            dd = nday // 4
            nday = (4 * dd + 3) // 1461
            dd = 4 * dd + 3 - 1461 * nday
            dd = (dd + 4) // 4
            mm = (5 * dd - 3) // 153
            dd = 5 * dd - 3 - 153 * mm
            dd = (dd + 5) // 5
            yyyy = 100 * century + nday;
            if mm < 10:
                mm += 3
            else:
                mm -= 9
                yyyy += 1

            ntime = bytes_to_int(v[4:])
            h = ntime // (3600 * ISC_TIME_SECONDS_PRECISION)
            ntime %= 3600 * ISC_TIME_SECONDS_PRECISION
            m = ntime // (60 * ISC_TIME_SECONDS_PRECISION)
            ntime %= 60 * ISC_TIME_SECONDS_PRECISION
            s = ntime // ISC_TIME_SECONDS_PRECISION
            ms = ntime % ISC_TIME_SECONDS_PRECISION * 100

            return datetime.datetime(yyyy, mm, dd, h, m, s, ms)
        else:
            return v

    def db_info(self, info_requests):
        DEBUG_OUTPUT("Connection::db_info()")
        if type(info_requests) == int:  # singleton
            r = self._db_info([info_requests])
            return self._db_info_convert_type(info_requests, r[0][1])
        else:
            results = {}
            rs = self._db_info(info_requests)
            for i in range(len(info_requests)):
                if rs[i][0] == isc_info_error:
                    results[info_requests[i]] = None
                else:
                    results[info_requests[i]] = self._db_info_convert_type(
                                                    info_requests[i], rs[i][1])
            return results

    def trans_info(self, info_requests):
        if self._transaction:
            return self._transaction.trans_info(info_requests)
        return {}

    def close(self):
        DEBUG_OUTPUT("Connection::close()")
        if self.sock is None:
            return
        if self.db_handle:
            if self.is_services:
                self._op_service_detach()
            else:
                self._op_detach()
            (h, oid, buf) = self._op_response()
        self.sock.close()
        self.sock = None
        self.db_handle = None

    def drop_database(self):
        DEBUG_OUTPUT("Connection::drop_database()")
        self._op_drop_database()
        (h, oid, buf) = self._op_response()
        self.sock.close()
        self.sock = None
        self.db_handle = None

    def event_conduit(self, event_names, timeout=None):
        return EventConduit(self, event_names, timeout)

    def __del__(self):
        if self.sock:
            self.close()

    def is_disconnect(self):
        return self.sock is None

class Transaction(object):
    def __init__(self, connection, is_autocommit=False):
        DEBUG_OUTPUT("Transaction::__init__()")
        self._connection = connection
        self._trans_handle = None
        self._autocommit = is_autocommit

    def _begin(self):
        DEBUG_OUTPUT("Transaction::_begin()")
        tpb = transaction_parameter_block[self.connection.isolation_level]
        if self._autocommit:
            tpb += bs([isc_tpb_autocommit])
        self.connection._op_transaction(tpb)
        (h, oid, buf) = self.connection._op_response()
        self._trans_handle = h
        self.is_dirty = False

    def begin(self):
        DEBUG_OUTPUT("Transaction::begin()")
        self._begin()

    def savepoint(self, name):
        if self._trans_handle is None:
            return
        self.connection._op_exec_immediate(self._trans_handle,
                        query='SAVEPOINT '+name)
        (h, oid, buf) = self.connection._op_response()

    def commit(self, retaining=False):
        DEBUG_OUTPUT("Transaction::commit()",
                        self._trans_handle, retaining)
        if self._trans_handle is None:
            return
        if not self.is_dirty:
            return
        if retaining:
            self.connection._op_commit_retaining(self._trans_handle)
            (h, oid, buf) = self.connection._op_response()
        else:
            self.connection._op_commit(self._trans_handle)
            (h, oid, buf) = self.connection._op_response()
            self._trans_handle = None
        self.is_dirty = False

    def rollback(self, retaining=False, savepoint=None):
        DEBUG_OUTPUT("Transaction::rollback()",
                        self._trans_handle, retaining, savepoint)
        if self._trans_handle is None:
            return
        if savepoint:
            self.connection._op_exec_immediate(self._trans_handle,
                        query='ROLLBACK TO '+savepoint)
            (h, oid, buf) = self.connection._op_response()
            return
        if not self.is_dirty:
            return
        if retaining:
            self.connection._op_rollback_retaining(self._trans_handle)
            (h, oid, buf) = self.connection._op_response()
        else:
            self.connection._op_rollback(self._trans_handle)
            (h, oid, buf) = self.connection._op_response()
            self._trans_handle = None
        self.is_dirty = False

    def _trans_info(self, info_requests):
        if info_requests[-1] == isc_info_end:
            self.connection._op_info_transaction(self.trans_handle,
                    bs(info_requests))
        else:
            self.connection._op_info_transaction(self.trans_handle,
                    bs(info_requests+type(info_requests)([isc_info_end])))
        (h, oid, buf) = self.connection._op_response()
        i = 0
        i_request = 0
        r = []
        while i < len(buf):
            req = byte_to_int(buf[i])
            if req == isc_info_end:
                break
            assert req == info_requests[i_request] or req == isc_info_error
            l = bytes_to_int(buf[i+1:i+3])
            r.append((req, buf[i+3:i+3+l]))
            i = i + 3 + l

            i_request += 1
        return r

    def trans_info(self, info_requests):
        if type(info_requests) == int:  # singleton
            r = self._trans_info([info_requests])
            return {info_requests: r[1][0]}
        else:
            results = {}
            rs = self._trans_info(info_requests)
            for i in range(len(info_requests)):
                if rs[i][0] == isc_info_tra_isolation:
                    v = (byte_to_int(rs[i][1][0]), byte_to_int(rs[i][1][1]))
                elif rs[i][0] == isc_info_error:
                    v = None
                else:
                    v = bytes_to_int(rs[i][1])
                results[info_requests[i]] = v
            return results

    @property
    def connection(self):
        return self._connection

    @property
    def trans_handle(self):
        if self._trans_handle is None:
            self._begin()
        return self._trans_handle


class RowMapping(Mapping):
    """dict like interface to result rows
    """
    __slots__ = ("_description", "_fields")

    def __init__(self, row, description):
        self._fields = fields = {}
        # result may contain multiple fields with the same name. The
        # RowMapping API ignores these additional fields.
        for i, descr in enumerate(description):
            fields.setdefault(descr[0], row[i])
        self._description = description

    def __getitem__(self, key):
        fields = self._fields
        # try unnormalized key first
        try:
            return fields[key]
        except KeyError:
            pass

        # normalize field name
        if key[0] == '"' and key[-1] == '"':
            # field names in quotes are case sensitive
            normkey = key[1:-1]
        else:
            # default is all upper case fields
            normkey = key.upper()

        try:
            return fields[normkey]
        except KeyError:
            raise KeyError("RowMapping has no field names '%s'. Available "
                           "field names are: %s" %
                           (key, ", ".join(self.keys())))

    def __iter__(self):
        return iter(self._fields)

    def __len__(self):
        return len(self._fields)

    def __repr__(self):
        fields = self._fields
        values = ["%s=%r" % (desc[0], fields[desc[0]])
                  for desc in self._description]
        return ("<RowMapping at 0x%08x with fields: %s>" %
                (id(self), ", ".join(values)))

